from igraph.drawing.utils import FakeModule
from typing import Any

__all__ = ("find_cairo",)
__docformat__ = "restructuredtext en"


def find_cairo() -> Any:
    """Tries to import the ``cairo`` Python module if it is installed,
    also trying ``cairocffi`` (a drop-in replacement of ``cairo``).
    Returns a fake module if everything fails.
    """
    module_names = ["cairo", "cairocffi"]
    module = FakeModule("Plotting not available; please install pycairo or cairocffi")
    for module_name in module_names:
        try:
            module = __import__(module_name)
            break
        except ImportError:
            pass
        except OSError:
            # cairocffi throws an OSError if it is installed but libcairo-2 is
            # not present on the system
            pass
    return module



from igraph._igraph import GraphBase
from igraph.clustering import VertexDendrogram, VertexClustering
from igraph.utils import deprecated

from typing import List, Sequence, Tuple


def _community_fastgreedy(graph, weights=None):
    """Community structure based on the greedy optimization of modularity.

    This algorithm merges individual nodes into communities in a way that
    greedily maximizes the modularity score of the graph. It can be proven
    that if no merge can increase the current modularity score, the
    algorithm can be stopped since no further increase can be achieved.

    This algorithm is said to run almost in linear time on sparse graphs.

    B{Reference}: A Clauset, MEJ Newman and C Moore: Finding community structure
    in very large networks. I{Phys Rev E} 70, 066111 (2004).

    @param weights: edge attribute name or a list containing edge
      weights
    @return: an appropriate L{VertexDendrogram} object.
    """
    merges, qs = GraphBase.community_fastgreedy(graph, weights)
    optimal_count = _optimal_cluster_count_from_merges_and_modularity(graph, merges, qs)
    return VertexDendrogram(
        graph, merges, optimal_count, modularity_params={"weights": weights}
    )


def _community_infomap(graph, edge_weights=None, vertex_weights=None, trials=10):
    """Finds the community structure of the network according to the Infomap
    method of Martin Rosvall and Carl T. Bergstrom.

    B{References}

      - M. Rosvall and C. T. Bergstrom: Maps of information flow reveal
        community structure in complex networks, I{PNAS} 105, 1118 (2008).
        U{https://doi.org/10.1073/pnas.0706851105},
        U{https://arxiv.org/abs/0707.0609}.
      - M. Rosvall, D. Axelsson, and C. T. Bergstrom: The map equation,
        I{Eur Phys. J Special Topics} 178, 13 (2009).
        U{https://doi.org/10.1140/epjst/e2010-01179-1},
        U{https://arxiv.org/abs/0906.1405}.

    @param edge_weights: name of an edge attribute or a list containing
      edge weights.
    @param vertex_weights: name of a vertex attribute or a list containing
      vertex weights.
    @param trials: the number of attempts to partition the network.
    @return: an appropriate L{VertexClustering} object with an extra attribute
      called C{codelength} that stores the code length determined by the
      algorithm.
    """
    membership, codelength = GraphBase.community_infomap(
        graph, edge_weights, vertex_weights, trials
    )
    return VertexClustering(
        graph,
        membership,
        params={"codelength": codelength},
        modularity_params={"weights": edge_weights},
    )


def _community_leading_eigenvector(
    graph, clusters=None, weights=None, arpack_options=None
):
    """Newman's leading eigenvector method for detecting community structure.

    This is the proper implementation of the recursive, divisive algorithm:
    each split is done by maximizing the modularity regarding the
    original network.

    B{Reference}: MEJ Newman: Finding community structure in networks using the
    eigenvectors of matrices, arXiv:physics/0605087

    @param clusters: the desired number of communities. If C{None}, the
      algorithm tries to do as many splits as possible. Note that the
      algorithm won't split a community further if the signs of the leading
      eigenvector are all the same, so the actual number of discovered
      communities can be less than the desired one.
    @param weights: name of an edge attribute or a list containing
      edge weights.
    @param arpack_options: an L{ARPACKOptions} object used to fine-tune
      the ARPACK eigenvector calculation. If omitted, the module-level
      variable called C{arpack_options} is used.
    @return: an appropriate L{VertexClustering} object.
    """
    if clusters is None:
        clusters = -1

    kwds = {"weights": weights}
    if arpack_options is not None:
        kwds["arpack_options"] = arpack_options

    membership, _, q = GraphBase.community_leading_eigenvector(graph, clusters, **kwds)
    return VertexClustering(graph, membership, modularity=q)


def _community_label_propagation(graph, weights=None, initial=None, fixed=None):
    """Finds the community structure of the graph according to the label
    propagation method of Raghavan et al.

    Initially, each vertex is assigned a different label. After that,
    each vertex chooses the dominant label in its neighbourhood in each
    iteration. Ties are broken randomly and the order in which the
    vertices are updated is randomized before every iteration. The
    algorithm ends when vertices reach a consensus.

    Note that since ties are broken randomly, there is no guarantee that
    the algorithm returns the same community structure after each run.
    In fact, they frequently differ. See the paper of Raghavan et al.
    on how to come up with an aggregated community structure.

    Also note that the community _labels_ (numbers) have no semantic meaning
    and igraph is free to re-number communities. If you use fixed labels,
    igraph may still re-number the communities, but co-community membership
    constraints will be respected: if you had two vertices with fixed labels
    that belonged to the same community, they will still be in the same
    community at the end. Similarly, if you had two vertices with fixed
    labels that belonged to different communities, they will still be in
    different communities at the end.

    B{Reference}: Raghavan, U.N. and Albert, R. and Kumara, S. Near linear
    time algorithm to detect community structures in large-scale networks.
    I{Phys Rev} E 76:036106, 2007. U{https://arxiv.org/abs/0709.2938}.

    @param weights: name of an edge attribute or a list containing
      edge weights
    @param initial: name of a vertex attribute or a list containing
      the initial vertex labels. Labels are identified by integers from
      zero to M{n-1} where M{n} is the number of vertices. Negative
      numbers may also be present in this vector, they represent unlabeled
      vertices.
    @param fixed: a list of booleans for each vertex. C{True} corresponds
      to vertices whose labeling should not change during the algorithm.
      It only makes sense if initial labels are also given. Unlabeled
      vertices cannot be fixed. It may also be the name of a vertex
      attribute; each attribute value will be interpreted as a Boolean.
    @return: an appropriate L{VertexClustering} object.
    """
    if isinstance(fixed, str):
        fixed = [bool(o) for o in graph.vs[fixed]]
    cl = GraphBase.community_label_propagation(graph, weights, initial, fixed)
    return VertexClustering(graph, cl, modularity_params={"weights": weights})


def _community_multilevel(graph, weights=None, return_levels=False, resolution=1):
    """Community structure based on the multilevel algorithm of
    Blondel et al.

    This is a bottom-up algorithm: initially every vertex belongs to a
    separate community, and vertices are moved between communities
    iteratively in a way that maximizes the vertices' local contribution
    to the overall modularity score. When a consensus is reached (i.e. no
    single move would increase the modularity score), every community in
    the original graph is shrunk to a single vertex (while keeping the
    total weight of the incident edges) and the process continues on the
    next level. The algorithm stops when it is not possible to increase
    the modularity anymore after shrinking the communities to vertices.

    This algorithm is said to run almost in linear time on sparse graphs.

    B{Reference}: VD Blondel, J-L Guillaume, R Lambiotte and E Lefebvre: Fast
    unfolding of community hierarchies in large networks, I{J Stat Mech}
    P10008 (2008). U{https://arxiv.org/abs/0803.0476}

    @param weights: edge attribute name or a list containing edge
      weights
    @param return_levels: if C{True}, the communities at each level are
      returned in a list. If C{False}, only the community structure with
      the best modularity is returned.
    @param resolution: the resolution parameter to use in the modularity
      measure. Smaller values result in a smaller number of larger clusters,
      while higher values yield a large number of small clusters. The classical
      modularity measure assumes a resolution parameter of 1.
    @return: a list of L{VertexClustering} objects, one corresponding to
      each level (if C{return_levels} is C{True}), or a L{VertexClustering}
      corresponding to the best modularity.
    """
    if graph.is_directed():
        raise ValueError("input graph must be undirected")

    modularity_params = {"weights": weights, "resolution": resolution}
    if return_levels:
        levels, qs = GraphBase.community_multilevel(
            graph, weights, return_levels=True, resolution=resolution
        )
        result = []
        for level, q in zip(levels, qs):
            result.append(
                VertexClustering(graph, level, q, modularity_params=modularity_params)
            )
    else:
        membership = GraphBase.community_multilevel(
            graph, weights, return_levels=False, resolution=resolution
        )
        result = VertexClustering(
            graph, membership, modularity_params=modularity_params
        )

    return result


def _community_optimal_modularity(graph, *args, **kwds):
    """Calculates the optimal modularity score of the graph and the
    corresponding community structure.

    This function uses the GNU Linear Programming Kit to solve a large
    integer optimization problem in order to find the optimal modularity
    score and the corresponding community structure, therefore it is
    unlikely to work for graphs larger than a few (less than a hundred)
    vertices. Consider using one of the heuristic approaches instead if
    you have such a large graph.

    @return: the calculated membership vector and the corresponding
      modularity in a tuple."""
    membership, modularity = GraphBase.community_optimal_modularity(
        graph, *args, **kwds
    )
    return VertexClustering(graph, membership, modularity)


def _community_edge_betweenness(graph, clusters=None, directed=True, weights=None):
    """Community structure based on the betweenness of the edges in the
    network.

    The idea is that the betweenness of the edges connecting two
    communities is typically high, as many of the shortest paths between
    nodes in separate communities go through them. So we gradually remove
    the edge with the highest betweenness and recalculate the betweennesses
    after every removal. This way sooner or later the network falls of to
    separate components. The result of the clustering will be represented
    by a dendrogram.

    When edge weights are given, the ratio of betweenness and weight values
    is used to choose which edges to remove first, as described in
    M. E. J. Newman: Analysis of Weighted Networks (2004), Section C.
    Thus, edges with large weights are treated as strong connections,
    and will be removed later than weak connections having similar betweenness.
    Weights are also used for calculating modularity.

    @param clusters: the number of clusters we would like to see. This
      practically defines the "level" where we "cut" the dendrogram to
      get the membership vector of the vertices. If C{None}, the dendrogram
      is cut at the level that maximizes the modularity when the graph is
      unweighted; otherwise the dendrogram is cut at at a single cluster
      (because cluster count selection based on modularities does not make
      sense for this method if not all the weights are equal).
    @param directed: whether the directionality of the edges should be
      taken into account or not.
    @param weights: name of an edge attribute or a list containing
      edge weights. Higher weights indicate stronger connections.
    @return: a L{VertexDendrogram} object, initally cut at the maximum
      modularity or at the desired number of clusters.
    """
    merges, qs = GraphBase.community_edge_betweenness(graph, directed, weights)
    if clusters is None:
        if qs is not None:
            clusters = _optimal_cluster_count_from_merges_and_modularity(
                graph, merges, qs
            )
        else:
            clusters = 1

    return VertexDendrogram(
        graph, merges, clusters, modularity_params={"weights": weights}
    )


def _community_spinglass(graph, *args, **kwds):
    """Finds the community structure of the graph according to the
    spinglass community detection method of Reichardt & Bornholdt.

    B{References}

      - Reichardt J and Bornholdt S: Statistical mechanics of community
        detection. I{Phys Rev E} 74:016110 (2006).
        U{https://arxiv.org/abs/cond-mat/0603718}.

      - Traag VA and Bruggeman J: Community detection in networks
        with positive and negative links. I{Phys Rev E} 80:036115 (2009).
        U{https://arxiv.org/abs/0811.2329}.

    @keyword weights: edge weights to be used. Can be a sequence or
      iterable or even an edge attribute name.
    @keyword spins: integer, the number of spins to use. This is the
      upper limit for the number of communities. It is not a problem
      to supply a (reasonably) big number here, in which case some
      spin states will be unpopulated.
    @keyword parupdate: whether to update the spins of the vertices in
      parallel (synchronously) or not
    @keyword start_temp: the starting temperature
    @keyword stop_temp: the stop temperature
    @keyword cool_fact: cooling factor for the simulated annealing
    @keyword update_rule: specifies the null model of the simulation.
      Possible values are C{"config"} (a random graph with the same
      vertex degrees as the input graph) or C{"simple"} (a random
      graph with the same number of edges)
    @keyword gamma: the gamma argument of the algorithm, specifying the
      balance between the importance of present and missing edges
      within a community. The default value of 1.0 assigns equal
      importance to both of them.
    @keyword implementation: currently igraph contains two implementations
      of the spinglass community detection algorithm. The faster
      original implementation is the default. The other implementation
      is able to take into account negative weights, this can be
      chosen by setting C{implementation} to C{"neg"}
    @keyword lambda_: the lambda argument of the algorithm, which
      specifies the balance between the importance of present and missing
      negatively weighted edges within a community. Smaller values of
      lambda lead to communities with less negative intra-connectivity.
      If the argument is zero, the algorithm reduces to a graph coloring
      algorithm, using the number of spins as colors. This argument is
      ignored if the original implementation is used. Note the underscore
      at the end of the argument name; this is due to the fact that
      lambda is a reserved keyword in Python.
    @return: an appropriate L{VertexClustering} object.
    """
    membership = GraphBase.community_spinglass(graph, *args, **kwds)
    if "weights" in kwds:
        modularity_params = {"weights": kwds["weights"]}
    else:
        modularity_params = {}
    return VertexClustering(graph, membership, modularity_params=modularity_params)


def _community_voronoi(graph, lengths=None, weights=None, mode="out", radius=None):
    """Finds communities using Voronoi partitioning.

    This function finds communities using a Voronoi partitioning of vertices based
    on the given edge lengths divided by the edge clustering coefficient.
    The generator vertices are chosen to be those with the largest local relative
    density within a radius, with the local relative density of a vertex defined
    as C{s * m / (m + k)}, where C{s} is the strength of the vertex, C{m} is
    the number of edges within the vertex's first order neighborhood, while C{k}
    is the number of edges with only one endpoint within this neighborhood.

    B{References}

      - Deritei et al., Community detection by graph Voronoi diagrams,
        I{New Journal of Physics} 16, 063007 (2014).
        U{https://doi.org/10.1088/1367-2630/16/6/063007}.
      - Molnár et al., Community Detection in Directed Weighted Networks using
        Voronoi Partitioning, I{Scientific Reports} 14, 8124 (2024).
        U{https://doi.org/10.1038/s41598-024-58624-4}.

    @param lengths: edge lengths, or C{None} to consider all edges as having
      unit length. Voronoi partitioning will use edge lengths equal to
      lengths / ECC where ECC is the edge clustering coefficient.
    @param weights: edge weights, or C{None} to consider all edges as having
      unit weight. Weights are used when selecting generator points, as well
      as for computing modularity.
    @param mode: specifies how to use the direction of edges when computing
      distances from generator points. If C{"out"} (the default), distances
      from generator points to all other nodes are considered following the
      direction of edges. If C{"in"}, distances are computed in the reverse
      direction (i.e., from all nodes to generator points). If C{"all"}, 
      edge directions are ignored and the graph is treated as undirected.
      This parameter is ignored for undirected graphs.
    @param radius: the radius/resolution to use when selecting generator points.
      The larger this value, the fewer partitions there will be. Pass C{None}
      to automatically select the radius that maximizes modularity.
    @return: an appropriate L{VertexClustering} object with an extra attribute
      called C{generators} (the generator vertices).
    """
    # Convert mode string to proper enum value to avoid deprecation warning
    if isinstance(mode, str):
        mode_map = {"out": "out", "in": "in", "all": "all", "total": "all"}  # alias
        if mode.lower() in mode_map:
            mode = mode_map[mode.lower()]
        else:
            raise ValueError(f"Invalid mode '{mode}'. Must be one of: out, in, all")
    
    membership, generators, modularity = GraphBase.community_voronoi(graph, lengths, weights, mode, radius)

    params = {"generators": generators}
    modularity_params = {}
    if weights is not None:
        modularity_params["weights"] = weights

    clustering = VertexClustering(
        graph, membership, modularity=modularity, params=params, modularity_params=modularity_params
    )

    clustering.generators = generators
    return clustering


def _community_walktrap(graph, weights=None, steps=4):
    """Community detection algorithm of Latapy & Pons, based on random
    walks.

    The basic idea of the algorithm is that short random walks tend to stay
    in the same community. The result of the clustering will be represented
    as a dendrogram.

    B{Reference}: Pascal Pons, Matthieu Latapy: Computing communities in large
    networks using random walks, U{https://arxiv.org/abs/physics/0512106}.

    @param weights: name of an edge attribute or a list containing
      edge weights
    @param steps: length of random walks to perform

    @return: a L{VertexDendrogram} object, initially cut at the maximum
      modularity.
    """
    merges, qs = GraphBase.community_walktrap(graph, weights, steps)
    optimal_count = _optimal_cluster_count_from_merges_and_modularity(graph, merges, qs)
    return VertexDendrogram(
        graph, merges, optimal_count, modularity_params={"weights": weights}
    )


def _k_core(graph, *args):
    """Returns some k-cores of the graph.

    The method accepts an arbitrary number of arguments representing
    the desired indices of the M{k}-cores to be returned. The arguments
    can also be lists or tuples. The result is a single L{Graph} object
    if an only integer argument was given, otherwise the result is a
    list of L{Graph} objects representing the desired k-cores in the
    order the arguments were specified. If no argument is given, returns
    all M{k}-cores in increasing order of M{k}.
    """
    if len(args) == 0:
        indices = range(graph.vcount())
        return_single = False
    else:
        return_single = True
        indices = []
        for arg in args:
            try:
                indices.extend(arg)
            except Exception:
                indices.append(arg)

    if len(indices) > 1 or hasattr(args[0], "__iter__"):
        return_single = False

    corenesses = graph.coreness()
    result = []
    vidxs = range(graph.vcount())
    for idx in indices:
        core_idxs = [vidx for vidx in vidxs if corenesses[vidx] >= idx]
        result.append(graph.subgraph(core_idxs))

    if return_single:
        return result[0]
    return result


def _community_leiden(
    graph,
    objective_function="CPM",
    weights=None,
    resolution=1.0,
    beta=0.01,
    initial_membership=None,
    n_iterations=2,
    node_weights=None,
    node_in_weights=None,
    **kwds,
):
    """Finds the community structure of the graph using the Leiden
    algorithm of Traag, van Eck & Waltman.

    B{Reference}: Traag, V. A., Waltman, L., & van Eck, N. J. (2019). From Louvain
    to Leiden: guaranteeing well-connected communities. I{Scientific Reports},
    9(1), 5233. doi: 10.1038/s41598-019-41695-z

    @param objective_function: whether to use the Constant Potts
      Model (CPM) or modularity. Must be either C{"CPM"} or C{"modularity"}.
    @param weights: edge weights to be used. Can be a sequence or
      iterable or even an edge attribute name.
    @param resolution: the resolution parameter to use. Higher resolutions
      lead to more smaller communities, while lower resolutions lead to fewer
      larger communities.
    @param beta: parameter affecting the randomness in the Leiden
      algorithm. This affects only the refinement step of the algorithm.
    @param initial_membership: if provided, the Leiden algorithm
      will try to improve this provided membership. If no argument is
      provided, the aglorithm simply starts from the singleton partition.
    @param n_iterations: the number of iterations to iterate the Leiden
      algorithm. Each iteration may improve the partition further. Using
      a negative number of iterations will run until a stable iteration is
      encountered (i.e. the quality was not increased during that
      iteration).
    @param node_weights: the node weights used in the Leiden algorithm.
      If this is not provided, it will be automatically determined on the
      basis of whether you want to use CPM or modularity. If you do provide
      this, please make sure that you understand what you are doing.
    @param node_in_weights: the inbound node weights used in the directed
      variant of the Leiden algorithm. If this is not provided, it will be
      automatically determined on the basis of whether you want to use CPM or
      modularity. If you do provide this, please make sure that you understand
      what you are doing.
    @return: an appropriate L{VertexClustering} object with an extra attribute
      called C{quality} that stores the value of the internal quality function
      optimized by the algorithm.
    """
    if objective_function.lower() not in ("cpm", "modularity"):
        raise ValueError('objective_function must be "CPM" or "modularity".')

    if "resolution_parameter" in kwds:
        deprecated(
            "resolution_parameter keyword argument is deprecated, use "
            "resolution=... instead"
        )
        resolution = kwds.pop("resolution_parameter")

    if kwds:
        raise TypeError("unexpected keyword argument")

    membership, quality = GraphBase.community_leiden(
        graph,
        edge_weights=weights,
        node_weights=node_weights,
        node_in_weights=node_in_weights,
        resolution=resolution,
        normalize_resolution=(objective_function == "modularity"),
        beta=beta,
        initial_membership=initial_membership,
        n_iterations=n_iterations,
    )

    params = {"quality": quality}

    modularity_params = {"resolution": resolution}
    if weights is not None:
        modularity_params["weights"] = weights

    return VertexClustering(
        graph, membership, params=params, modularity_params=modularity_params
    )


def _community_fluid_communities(graph, no_of_communities):
    """Community detection based on fluids interacting on the graph.

    The algorithm is based on the simple idea of several fluids interacting 
    in a non-homogeneous environment (the graph topology), expanding and 
    contracting based on their interaction and density. Weighted graphs are 
    not supported.

    This function implements the community detection method described in:
    Parés F, Gasulla DG, et. al. (2018) Fluid Communities: A Competitive,
    Scalable and Diverse Community Detection Algorithm.

    @param no_of_communities: The number of communities to be found. Must be
      greater than 0 and fewer than or equal to the number of vertices in the graph.
    @return: an appropriate L{VertexClustering} object.
    """
    # Validate input parameters
    if no_of_communities <= 0:
        raise ValueError("no_of_communities must be greater than 0")
    
    if no_of_communities > graph.vcount():
        raise ValueError("no_of_communities must be fewer than or equal to the number of vertices")
    
    # Check if graph is weighted (not supported)
    if graph.is_weighted():
        raise ValueError("Weighted graphs are not supported by the fluid communities algorithm")
    
    # Handle directed graphs - the algorithm works on undirected graphs
    # but can accept directed graphs (they are treated as undirected)
    if graph.is_directed():
        import warnings
        warnings.warn(
            "Directed graphs are treated as undirected in the fluid communities algorithm",
            UserWarning,
            stacklevel=2
        )
    
    membership = GraphBase.community_fluid_communities(graph, no_of_communities)
    return VertexClustering(graph, membership)
  
  
def _modularity(self, membership, weights=None, resolution=1, directed=True):
    """Calculates the modularity score of the graph with respect to a given
    clustering.

    The modularity of a graph w.r.t. some division measures how good the
    division is, or how separated are the different vertex types from each
    other. It's defined as M{Q=1/(2m)*sum(Aij-gamma*ki*kj/(2m)delta(ci,cj),i,j)}.
    M{m} is the number of edges, M{Aij} is the element of the M{A}
    adjacency matrix in row M{i} and column M{j}, M{ki} is the degree of
    node M{i}, M{kj} is the degree of node M{j}, and M{Ci} and C{cj} are
    the types of the two vertices (M{i} and M{j}), and M{gamma} is a resolution
    parameter that defaults to 1 for the classical definition of modularity.
    M{delta(x,y)} is one iff M{x=y}, 0 otherwise.

    If edge weights are given, the definition of modularity is modified as
    follows: M{Aij} becomes the weight of the corresponding edge, M{ki}
    is the total weight of edges adjacent to vertex M{i}, M{kj} is the
    total weight of edges adjacent to vertex M{j} and M{m} is the total
    edge weight in the graph.

    B{Reference}: MEJ Newman and M Girvan: Finding and evaluating community
    structure in networks. I{Phys Rev E} 69 026113, 2004.

    @param membership: a membership list or a L{VertexClustering} object
    @param weights: optional edge weights or C{None} if all edges are
      weighed equally. Attribute names are also allowed.
    @param resolution: the resolution parameter I{gamma} in the formula above.
      The classical definition of modularity is retrieved when the resolution
      parameter is set to 1.
    @param directed: whether to consider edge directions if the graph is directed.
      C{True} will use the directed variant of the modularity measure where the
      in- and out-degrees of nodes are treated separately; C{False} will treat
      directed graphs as undirected.
    @return: the modularity score
    """
    if isinstance(membership, VertexClustering):
        if membership.graph != self:
            raise ValueError("clustering object belongs to another graph")
        return GraphBase.modularity(
            self, membership.membership, weights, resolution, directed
        )
    else:
        return GraphBase.modularity(self, membership, weights, resolution, directed)


def _optimal_cluster_count_from_merges_and_modularity(
    graph, merges: Sequence[Tuple[int, int]], qs: List[float]
) -> float:
    """Helper function to find the optimal cluster count for a hierarchical
    clustering of a graph, given the merge matrix and the list of modularity
    values after each merge.

    Reverses the modularity vector as a side effect.
    """
    no_of_comps = graph.vcount() - len(merges)
    qs.reverse()
    return qs.index(max(qs)) + no_of_comps



"""
.. _tutorials-bridges:

========
Bridges
========

This example shows how to compute and visualize the `bridges <https://en.wikipedia.org/wiki/Bridge_(graph_theory)>`_ in a graph using :meth:`igraph.GraphBase.bridges`. For an example on articulation points instead, see :ref:`tutorials-articulation-points`.
"""

import igraph as ig
import matplotlib.pyplot as plt

# %%
# Let's start with a simple example. We begin by constructing a graph that
# includes a few bridges:
g = ig.Graph(14, [(0, 1), (1, 2), (2, 3), (0, 3), (0, 2), (1, 3), (3, 4),
        (4, 5), (5, 6), (6, 4), (6, 7), (7, 8), (7, 9), (9, 10), (10 ,11),
        (11 ,7), (7, 10), (8, 9), (8, 10), (5, 12), (12, 13)])

# %%
# Then we can use a function to actually find the bridges, i.e. the edges that
# connect different parts of the graph:
bridges = g.bridges()

# %%
# We set a separate color for those edges, to emphasize then in a plot:
g.es["color"] = "gray"
g.es[bridges]["color"] = "red"
g.es["width"] = 0.8
g.es[bridges]["width"] = 1.2

# %%
# Finally, we plot the graph using that emphasis:
fig, ax = plt.subplots()
ig.plot(
    g,
    target=ax,
    vertex_size=30,
    vertex_color="lightblue",
    vertex_label=range(g.vcount()),
)
plt.show()

# %%
# Advanced: Cutting Effect
# --------------------------
# Bridges are edges that when removed, will separate the graph into more components then they started with. To emphasise the removal of edges from the graph, we can add small "x" effect to each of the bridges by using edge labels.

# %%
# As before, we begin by constructing the graph:
g = ig.Graph(14, [(0, 1), (1, 2), (2, 3), (0, 3), (0, 2), (1, 3), (3, 4),
        (4, 5), (5, 6), (6, 4), (6, 7), (7, 8), (7, 9), (9, 10), (10 ,11),
        (11 ,7), (7, 10), (8, 9), (8, 10), (5, 12), (12, 13)])

# %%
# We then find and set the color for the bridges, but this time we also set a
# label for those edges:
bridges = g.bridges()
g.es["color"] = "gray"
g.es[bridges]["color"] = "red"
g.es["width"] = 0.8
g.es[bridges]["width"] = 1.2
g.es["label"] = ""
g.es[bridges]["label"] = "x"

# %%
# Finally, we can plot the graph:
fig, ax = plt.subplots()
ig.plot(
    g,
    target=ax,
    vertex_size=30,
    vertex_color="lightblue",
    vertex_label=range(g.vcount()),
    edge_background="#FFF0",  # transparent background color
    edge_align_label=True,  # make sure labels are aligned with the edge
    edge_label=g.es["label"],
    edge_label_color="red",
)
plt.show()



"""
.. _tutorials-stochastic-variability:

=========================================================
Stochastic Variability in Community Detection Algorithms
=========================================================

This example demonstrates the use of stochastic community detection methods to check whether a network possesses a strong community structure, and whether the partitionings we obtain are meaningul. Many community detection algorithms are randomized, and return somewhat different results after each run, depending on the random seed that was set. When there is a robust community structure, we expect these results to be similar to each other. When the community structure is weak or non-existent, the results may be noisy and highly variable. We will employ several partion similarity measures to analyse the consistency of the results, including the normalized mutual information (NMI), the variation of information (VI), and the Rand index (RI).

"""

# %%
import igraph as ig
import matplotlib.pyplot as plt
import itertools
import random

# %%
# .. note::
#   We set a random seed to ensure that the results look exactly the same in
#   the gallery. You don't need to do this when exploring randomness.
random.seed(42)

# %%
# We will use Zachary's karate club dataset [1]_, a classic example of a network
# with a strong community structure:
karate = ig.Graph.Famous("Zachary")

# %%
# We will compare it to an an Erdős-Rényi :math:`G(n, m)` random network having
# the same number of vertices and edges. The parameters 'n' and 'm' refer to the
# vertex and edge count, respectively. Since this is a random network, it should
# have no community structure.
random_graph = ig.Graph.Erdos_Renyi(n=karate.vcount(), m=karate.ecount())

# %%
# First, let us plot the two networks for a visual comparison:

# Create subplots
fig, axes = plt.subplots(1, 2, figsize=(12, 6), subplot_kw={"aspect": "equal"})

# Karate club network
ig.plot(
    karate,
    target=axes[0],
    vertex_color="lightblue",
    vertex_size=30,
    vertex_label=range(karate.vcount()),
    vertex_label_size=10,
    edge_width=1,
)
axes[0].set_title("Karate club network")

# Random network
ig.plot(
    random_graph,
    target=axes[1],
    vertex_color="lightcoral",
    vertex_size=30,
    vertex_label=range(random_graph.vcount()),
    vertex_label_size=10,
    edge_width=1,
)
axes[1].set_title("Erdős-Rényi random network")

plt.show()


# %%
# Function to compute similarity between partitions using various methods:
def compute_pairwise_similarity(partitions, method):
    similarities = []

    for p1, p2 in itertools.combinations(partitions, 2):
        similarity = ig.compare_communities(p1, p2, method=method)
        similarities.append(similarity)

    return similarities


# %%
# The Leiden method, accessible through :meth:`igraph.Graph.community_leiden()`,
# is a modularity maximization approach for community detection.  Since exact
# modularity maximization is NP-hard, the algorithm employs a greedy heuristic
# that processes vertices in a random order.  This randomness leads to
# variation in the detected communities across different runs, which is why
# results may differ each time the method is applied. The following function
# runs the Leiden algorithm multiple times:
def run_experiment(graph, iterations=100):
    partitions = [
        graph.community_leiden(objective_function="modularity").membership
        for _ in range(iterations)
    ]
    nmi_scores = compute_pairwise_similarity(partitions, method="nmi")
    vi_scores = compute_pairwise_similarity(partitions, method="vi")
    ri_scores = compute_pairwise_similarity(partitions, method="rand")
    return nmi_scores, vi_scores, ri_scores


# %%
# Run the experiment on both networks:
nmi_karate, vi_karate, ri_karate = run_experiment(karate)
nmi_random, vi_random, ri_random = run_experiment(random_graph)

# %%
# Finally, let us plot histograms of the pairwise similarities of the obtained
# partitionings to understand the result:
fig, axes = plt.subplots(2, 3, figsize=(12, 6))
measures = [
    # Normalized Mutual Information (0-1, higher = more similar)
    (nmi_karate, nmi_random, "NMI", 0, 1),
    # Variation of Information (0+, lower = more similar)
    (vi_karate, vi_random, "VI", 0, max(vi_karate + vi_random)),
    # Rand Index (0-1, higher = more similar)
    (ri_karate, ri_random, "RI", 0, 1),
]
colors = ["red", "blue", "green"]

for i, (karate_scores, random_scores, measure, lower, upper) in enumerate(measures):
    # Karate club histogram
    axes[0][i].hist(
        karate_scores,
        bins=20,
        range=(lower, upper),
        density=True,  # Probability density
        alpha=0.7,
        color=colors[i],
        edgecolor="black",
    )
    axes[0][i].set_title(f"{measure} - Karate club network")
    axes[0][i].set_xlabel(f"{measure} score")
    axes[0][i].set_ylabel("PDF")

    # Random network histogram
    axes[1][i].hist(
        random_scores,
        bins=20,
        range=(lower, upper),
        density=True,
        alpha=0.7,
        color=colors[i],
        edgecolor="black",
    )
    axes[1][i].set_title(f"{measure} - Random network")
    axes[1][i].set_xlabel(f"{measure} score")
    axes[0][i].set_ylabel("PDF")

plt.tight_layout()
plt.show()

# %%
# We have compared the pairwise similarities using the NMI, VI, and RI measures
# between partitonings obtained for the karate club network (strong community
# structure) and a comparable random graph (which lacks communities).
#
# The Normalized Mutual Information (NMI) and Rand Index (RI) both quantify
# similarity, and take values from :math:`[0,1]`. Higher values indicate more
# similar partitionings, with a value of 1 attained when the partitionings are
# identical.
#
# The Variation of Information (VI) is a distance measure. It takes values from
# :math:`[0,\infty]`, with lower values indicating higher similarities. Identical
# partitionings have a distance of zero.
#
# For the karate club network, NMI and RI value are concentrated near 1, while
# VI is concentrated near 0, suggesting a robust community structure. In contrast
# the values obtained for the random network are much more spread out, showing
# inconsistent partitionings due to the lack of a clear community structure.

# %%
# .. [1] W. Zachary: "An Information Flow Model for Conflict and Fission in Small Groups". Journal of Anthropological Research 33, no. 4 (1977): 452–73. https://www.jstor.org/stable/3629752



"""
Drawing and plotting routines for igraph.

igraph has two stable plotting backends at the moment: Cairo and Matplotlib.
It also has experimental support for plotly.

The Cairo backend is dependent on the C{pycairo} or C{cairocffi} libraries that
provide Python bindings to the popular U{Cairo library<https://www.cairographics.org>}.
This means that if you don't have U{pycairo<https://pycairo.readthedocs.io/>}
or U{cairocffi<https://doc.courtbouillon.org/cairocffi/>} installed, you won't be able
to use the Cairo plotting backend. Whenever the documentation refers to the
C{pycairo} library, you can safely replace it with C{cairocffi} as the two are
API-compatible.

The Matplotlib backend uses the U{Matplotlib library<https://matplotlib.org>}.
You will need to install it from PyPI if you want to use the Matplotlib
plotting backend. Many of our gallery examples use the matplotlib backend.

The plotly backend uses the U{plotly library <https://plotly.com/python/>} and,
like matplotlib, requires installation from PyPI.

If you do not want to (or cannot) install any of the dependencies outlined
above, you can still save the graph to an SVG file and view it from
U{Mozilla Firefox<https://www.mozilla.org/firefox>} (free) or edit it in
U{Inkscape<https://www.inkscape.org>} (free), U{Skencil<https://www.skencil.org>}
(formerly known as Sketch, also free) or Adobe Illustrator.
"""

from pathlib import Path
from warnings import warn

from igraph.configuration import Configuration
from igraph.drawing.cairo.utils import find_cairo
from igraph.drawing.matplotlib.utils import find_matplotlib
from igraph.drawing.plotly.utils import find_plotly
from igraph.drawing.cairo.plot import CairoPlot
from igraph.drawing.colors import Palette, palettes

from igraph.drawing.cairo.graph import CairoGraphDrawer
from igraph.drawing.cairo.matrix import CairoMatrixDrawer
from igraph.drawing.cairo.histogram import CairoHistogramDrawer
from igraph.drawing.cairo.palette import CairoPaletteDrawer
from igraph.drawing.matplotlib.graph import MatplotlibGraphDrawer
from igraph.drawing.matplotlib.matrix import MatplotlibMatrixDrawer
from igraph.drawing.matplotlib.histogram import MatplotlibHistogramDrawer
from igraph.drawing.matplotlib.palette import MatplotlibPaletteDrawer
from igraph.drawing.plotly.graph import PlotlyGraphDrawer

from igraph.drawing.utils import BoundingBox, Point, Rectangle
from igraph.utils import _is_running_in_ipython

__all__ = (
    "BoundingBox",
    "CairoGraphDrawer",
    "MatplotlibGraphDrawer",
    "DefaultGraphDrawer",
    "Plot",
    "Point",
    "Rectangle",
    "plot",
    "DrawerDirectory",
)

# TODO: deprecate
Plot = CairoPlot

# TODO: deprecate
DefaultGraphDrawer = CairoGraphDrawer


class DrawerDirectory:
    """Static class that finds the object/backend drawer

    This directory is used by the __plot__ functions.
    """

    valid_backends = ("cairo", "matplotlib")
    valid_objects = (
        "Graph",
        "Matrix",
        "Histogram",
        "Palette",
    )
    known_drawers = {
        "cairo": {
            "Graph": CairoGraphDrawer,
            "Matrix": CairoMatrixDrawer,
            "Histogram": CairoHistogramDrawer,
            "Palette": CairoPaletteDrawer,
        },
        "matplotlib": {
            "Graph": MatplotlibGraphDrawer,
            "Matrix": MatplotlibMatrixDrawer,
            "Histogram": MatplotlibHistogramDrawer,
            "Palette": MatplotlibPaletteDrawer,
        },
        "plotly": {
            "Graph": PlotlyGraphDrawer,
        },
    }

    @classmethod
    def resolve(cls, obj, backend):
        """Given a shape name, returns the corresponding shape drawer class

        @param cls: the class to resolve
        @param obj: an instance of the object to plot
        @param backend: the name of the backend
        @return: the corresponding shape drawer class

        @raise ValueError: if no drawer is available for this backend/object
        """
        object_name = str(obj.__class__).split(".")[-1].strip("<'>")

        try:
            return cls.known_drawers[backend][object_name]
        except KeyError:
            raise ValueError(
                f"unknown drawer for {object_name} and backend {backend}",
            ) from None


def plot(obj, target=None, bbox=(0, 0, 600, 600), *args, **kwds):
    """Plots the given object to the given target.

    Positional and keyword arguments not explicitly mentioned here will be
    passed down to the C{__plot__} method of the object being plotted.
    Since you are most likely interested in the keyword arguments available
    for graph plots, see L{Graph.__plot__} as well.

    @param obj: the object to be plotted
    @param target: the target where the object should be plotted. It can be one
      of the following types:

        - C{matplotib.axes.Axes} -- a matplotlib/pyplot axes in which the
          graph will be plotted. Drawing is delegated to the chosen matplotlib
          backend, and you can use interactive backends and matplotlib
          functions to save to file as well.

        - C{string} -- a file with the given name will be created and the plot
          will be stored there. If you are using the Cairo backend, an
          appropriate Cairo surface will be attached to the file. If you are
          using the matplotlib backend, the Figure will be saved to that file
          using Figure.savefig with default parameters. The supported image
          formats for Cairo are: PNG, PDF, SVG and PostScript; matplotlib might
          support additional formats.

        - C{cairo.Surface} -- the given Cairo surface will be used. This can
          refer to a PNG image, an arbitrary window, an SVG file, anything that
          Cairo can handle.

        - C{None} -- If you are using the Cairo backend, no plotting will be
          performed; igraph simply returns a CairoPlot_ object that you can use
          to manipulate the plot and save it to a file later. If you are using
          the matplotlib backend, a Figure objet and an Axes are created and
          the Axes is returned so you can manipulate it further. Similarly, if
          you are using the plotly backend, a Figure object is returned.

    @param bbox: the bounding box of the plot. It must be a tuple with either
      two or four integers, or a L{BoundingBox} object. If this is a tuple
      with two integers, it is interpreted as the width and height of the plot
      (in pixels for PNG images and on-screen plots, or in points for PDF,
      SVG and PostScript plots, where 72 pt = 1 inch = 2.54 cm). If this is
      a tuple with four integers, the first two denotes the X and Y coordinates
      of a corner and the latter two denoting the X and Y coordinates of the
      opposite corner. Ignored for Matplotlib plots.

    @keyword opacity: the opacity of the object being plotted. It can be
      used to overlap several plots of the same graph if you use the same
      layout for them -- for instance, you might plot a graph with opacity
      0.5 and then plot its spanning tree over it with opacity 0.1. To
      achieve this, you'll need to modify the L{Plot} object returned with
      L{Plot.add}. Ignored for Matplotlib plots.

    @keyword palette: the palette primarily used on the plot if the
      added objects do not specify a private palette. Must be either
      an L{igraph.drawing.colors.Palette} object or a string referring
      to a valid key of C{igraph.drawing.colors.palettes} (see module
      L{igraph.drawing.colors}) or C{None}. In the latter case, the default
      palette given by the configuration key C{plotting.palette} is used.

    @keyword margin: the top, right, bottom, left margins as a 4-tuple.
      If it has less than 4 elements or is a single float, the elements
      will be re-used until the length is at least 4. The default margin
      is 20 units on each side. Ignored for Matplotlib plots.

    @keyword inline: whether to try and show the plot object inline in the
      current IPython notebook. Passing C{None} here or omitting this keyword
      argument will look up the preferred behaviour from the
      C{shell.ipython.inlining.Plot} configuration key.  Note that this keyword
      argument has an effect only if igraph is run inside IPython and C{target}
      is C{None}.

    @keyword backend: the plotting backend to use; one of C{"cairo"},
      C{"matplotlib"} or C{"plotly"}. C{None} means to try to decide the backend
      from the plotting target and the default igraph configuration object.

    @return: an appropriate L{CairoPlot} object for the Cairo backend, the
      Matplotlib C{Axes} object for the Matplotlib backend, and the C{Figure}
      object for the plotly backend.

    @see: Graph.__plot__
    """

    VALID_BACKENDS = ("cairo", "matplotlib", "plotly")

    _, plt = find_matplotlib()
    cairo = find_cairo()
    plotly = find_plotly()

    backend = kwds.pop("backend", None)

    # Switch backend based on target (first) and config (second) if it was not
    # selected explicitly
    if backend is not None:
        pass
    elif hasattr(plt, "Axes") and isinstance(target, plt.Axes):
        backend = "matplotlib"
    elif hasattr(plotly, "graph_objects") and isinstance(
        target, plotly.graph_objects.Figure
    ):
        backend = "plotly"
    elif hasattr(cairo, "Surface") and isinstance(target, cairo.Surface):
        backend = "cairo"
    else:
        backend = Configuration.instance()["plotting.backend"]

    if backend not in VALID_BACKENDS:
        raise ValueError(f"unknown plotting backend: {backend!r}")

    if backend in ("matplotlib", "plotly"):
        # Choose palette
        # If explicit, use it. If not or None, ask the object: None is an
        # acceptable response from the object (e.g. for clusterings), it means
        # the palette is handled internally. If no response, default to config.
        palette = kwds.pop("palette", None)
        if palette is None:
            palette = getattr(
                obj,
                "_default_palette",
                Configuration.instance()["plotting.palette"],
            )
        if palette is not None and not isinstance(palette, Palette):
            palette = palettes[palette]

        if isinstance(target, (str, Path)):
            save_path = str(target)
            target = None
        else:
            save_path = None

        if target is None:
            if backend == "matplotlib":
                # Use get current axes, customary in these cases
                target = plt.gca()
            elif backend == "plotly":
                # Create a new figure if needed
                target = plotly.graph_objects.Figure()

        # Get the plotting function from the object
        plotter = getattr(obj, "__plot__", None)
        if plotter is None:
            warn("%s does not support plotting" % (obj,), stacklevel=1)
            return
        else:
            result = plotter(
                backend,
                target,
                palette=palette,
                *args,  # noqa: B026
                **kwds,
            )

            if save_path is not None:
                if backend == "matplotlib":
                    target.figure.savefig(save_path)
                elif backend == "plotly":
                    target.write_image(save_path)

            # For matplotlib, return the container artist, which makes it easier
            # to manipulate post-facto. The user can always get the artist with
            # plt.gca() - as we do, in fact.
            if backend == "matplotlib":
                return result

            return target

    # Cairo backend
    inline = False
    if target is None and _is_running_in_ipython():
        inline = kwds.get("inline")
        if inline is None:
            inline = Configuration.instance()["shell.ipython.inlining.Plot"]

    palette = kwds.pop("palette", None)
    background = kwds.pop("background", "white")
    margin = float(kwds.pop("margin", 20))
    result = CairoPlot(
        target=target,
        bbox=bbox,
        palette=palette,
        background=background,
    )
    item_bbox = result.bbox.contract(margin)
    result.add(obj, item_bbox, *args, **kwds)

    # If we requested an inline plot, just return the result and IPython will
    # call its _repr_svg_ method. If we requested a non-inline plot, show the
    # plot in a separate window and return nothing
    if inline:
        return result

    # We are either not in IPython or the user specified an explicit plot target,
    # so just show or save the result

    if isinstance(target, (str, Path)):
        # save
        result.save()

    # Also return the plot itself
    return result



"""
.. _tutorials-minimum-spanning-trees:

======================
Minimum Spanning Trees
======================

This example shows how to generate a `minimum spanning tree <https://en.wikipedia.org/wiki/Minimum_spanning_tree>`_ from an input graph using :meth:`igraph.Graph.spanning_tree`. If you only need a regular spanning tree, check out :ref:`tutorials-spanning-trees`.

"""

import random
import igraph as ig
import matplotlib.pyplot as plt

# %%
# We start by generating a grid graph with random integer weights between 1 and
# 20:
random.seed(0)
g = ig.Graph.Lattice([5, 5], circular=False)
g.es["weight"] = [random.randint(1, 20) for _ in g.es]

# %%
# We can then compute a minimum spanning tree using
# :meth:`igraph.Graph.spanning_tree`, making sure to pass in the randomly
# generated weights.
mst_edges = g.spanning_tree(weights=g.es["weight"], return_tree=False)

# %%
# We can print out the minimum edge weight sum
print("Minimum edge weight sum:", sum(g.es[mst_edges]["weight"]))

# Minimum edge weight sum: 136

# %%
# Finally, we can plot the graph, highlighting the edges that are part of the
# minimum spanning tree.
g.es["color"] = "lightgray"
g.es[mst_edges]["color"] = "midnightblue"
g.es["width"] = 1.0
g.es[mst_edges]["width"] = 3.0

fig, ax = plt.subplots()
ig.plot(
    g,
    target=ax,
    layout="grid",
    vertex_color="lightblue",
    edge_width=g.es["width"],
    edge_label=g.es["weight"],
    edge_background="white",
)
plt.show()



"""
.. _tutorials-bipartite-matching-maxflow:

==========================================
Maximum Bipartite Matching by Maximum Flow
==========================================

This example presents how to visualise bipartite matching using maximum flow (see :meth:`igraph.Graph.maxflow`).

.. note::  :meth:`igraph.Graph.maximum_bipartite_matching` is usually a better way to find the maximum bipartite matching. For a demonstration on how to use that method instead, check out :ref:`tutorials-bipartite-matching`.
"""

import igraph as ig
import matplotlib.pyplot as plt

# %%
# We start by creating the bipartite directed graph.
g = ig.Graph(
    9,
    [(0, 4), (0, 5), (1, 4), (1, 6), (1, 7), (2, 5), (2, 7), (2, 8), (3, 6), (3, 7)],
    directed=True,
)

# %%
# We assign:
#  - nodes 0-3 to one side
#  - nodes 4-8 to the other side
g.vs[range(4)]["type"] = True
g.vs[range(4, 9)]["type"] = False

# %%
# Then we add a source (vertex 9) and a sink (vertex 10)
g.add_vertices(2)
g.add_edges([(9, 0), (9, 1), (9, 2), (9, 3)])  # connect source to one side
g.add_edges([(4, 10), (5, 10), (6, 10), (7, 10), (8, 10)])  # ... and sinks to the other

flow = g.maxflow(9, 10)
print("Size of maximum matching (maxflow) is:", flow.value)

# %%
# Let's compare the output against :meth:`igraph.Graph.maximum_bipartite_matching`:

# delete the source and sink, which are unneeded for this function.
g2 = g.copy()
g2.delete_vertices([9, 10])
matching = g2.maximum_bipartite_matching()
matching_size = sum(1 for i in range(4) if matching.is_matched(i))
print("Size of maximum matching (maximum_bipartite_matching) is:", matching_size)

# %%
# Last, we can display the original flow graph nicely with the matchings added.
# To achieve a pleasant visual effect, we set the positions of source and sink
# manually:
layout = g.layout_bipartite()
layout[9] = (2, -1)
layout[10] = (2, 2)

fig, ax = plt.subplots()
ig.plot(
    g,
    target=ax,
    layout=layout,
    vertex_size=30,
    vertex_label=range(g.vcount()),
    vertex_color=["lightblue" if i < 9 else "orange" for i in range(11)],
    edge_width=[1.0 + flow.flow[i] for i in range(g.ecount())],
)
plt.show()



# vim:ts=4:sw=4:sts=4:et
# -*- coding: utf-8 -*-
"""
Layout-related code in the igraph library.

This package contains the implementation of the L{Layout} object.
"""

from math import sin, cos, pi

from igraph._igraph import GraphBase
from igraph.drawing.utils import BoundingBox
from igraph.statistics import RunningMean


__all__ = (
    "Layout",
    "align_layout",
    "_layout",
    "_layout_auto",
    "_layout_sugiyama",
    "_layout_method_wrapper",
    "_3d_version_for",
    "_layout_mapping",
)


class Layout:
    """Represents the layout of a graph.

    A layout is practically a list of coordinates in an n-dimensional
    space. This class is generic in the sense that it can store coordinates
    in any n-dimensional space.

    Layout objects are not associated directly with a graph. This is deliberate:
    there were times when I worked with almost identical copies of the same
    graph, the only difference was that they had different colors assigned to
    the vertices. It was particularly convenient for me to use the same layout
    for all of them, especially when I made figures for a paper. However,
    C{igraph} will of course refuse to draw a graph with a layout that has
    fewer coordinates than the node count of the graph.

    Layouts behave exactly like lists when they are accessed using the item
    index operator (C{[...]}). They can even be iterated through. Items
    returned by the index operator are only copies of the coordinates,
    but the stored coordinates can be modified by directly assigning to
    an index.

        >>> layout = Layout([(0, 1), (0, 2)])
        >>> coords = layout[1]
        >>> print(coords)
        [0, 2]
        >>> coords = (0, 3)
        >>> print(layout[1])
        [0, 2]
        >>> layout[1] = coords
        >>> print(layout[1])
        [0, 3]

    Optionally, a layout may have I{edge routing} information attached to
    each edge in the layout in the L{edge_routing} property.

    @ivar edge_routing: C{None} if no edge routing information is available,
    or a list of lists of control point coordinates, one for each edge. When
    an edge has control points, it should be drawn in a way that the edge
    passes through all the control points in the order they appear in the list.
    """

    def __init__(self, coords=None, dim=None):
        """Constructor.

        @param coords: the coordinates to be stored in the layout.
        @param dim: the number of dimensions. If C{None}, the number of
        dimensions is determined automatically from the length of the first
        item of the coordinate list. If there are no entries in the coordinate
        list, the default will be 2.  Generally, this should be given if the
        length of the coordinate list is zero, otherwise it should be left as
        is.
        """
        self.edge_routing = None

        if coords is not None:
            self._coords = [list(coord) for coord in coords]
        else:
            self._coords = []

        if dim is None:
            if len(self._coords) == 0:
                self._dim = 2
            else:
                self._dim = len(self._coords[0])
        else:
            self._dim = int(dim)
            for row in self._coords:
                if len(row) != self._dim:
                    raise ValueError(
                        "all items in the coordinate list "
                        + "must have a length of %d" % self._dim
                    )

    def __len__(self):
        return len(self._coords)

    def __getitem__(self, idx):
        return self._coords[idx]

    def __setitem__(self, idx, value):
        if len(value) != self._dim:
            raise ValueError("assigned item must have %d elements" % self._dim)
        self._coords[idx] = list(value)

    def __delitem__(self, idx):
        del self._coords[idx]

    def __copy__(self):
        return self.__class__(self.coords, self.dim)

    def __repr__(self):
        if not self.coords:
            vertex_count = "no vertices"
        elif len(self.coords) == 1:
            vertex_count = "1 vertex"
        else:
            vertex_count = "%d vertices" % len(self.coords)
        if self.dim == 1:
            dim_count = "1 dimension"
        else:
            dim_count = "%d dimensions" % self.dim
        return "<%s with %s and %s>" % (
            self.__class__.__name__,
            vertex_count,
            dim_count,
        )

    @property
    def dim(self):
        """Returns the number of dimensions"""
        return self._dim

    @property
    def coords(self):
        """The coordinates as a list of lists"""
        return [row[:] for row in self._coords]

    def append(self, value):
        """Appends a new point to the layout"""
        if len(value) < self._dim:
            raise ValueError("appended item must have %d elements" % self._dim)
        self._coords.append([float(coord) for coord in value[0 : self._dim]])

    def mirror(self, dim):
        """Mirrors the layout along the given dimension(s)

        @param dim: the list of dimensions or a single dimension
        """
        if isinstance(dim, int):
            dim = [dim]
        else:
            dim = [int(x) for x in dim]

        for current_dim in dim:
            for row in self._coords:
                row[current_dim] *= -1

    def rotate(self, angle, dim1=0, dim2=1, **kwds):
        """Rotates the layout by the given degrees on the plane defined by
        the given two dimensions.

        @param angle: the angle of the rotation, specified in degrees.
        @param dim1: the first axis of the plane of the rotation.
        @param dim2: the second axis of the plane of the rotation.
        @keyword origin: the origin of the rotation. If not specified, the
          origin will be the origin of the coordinate system.
        """

        origin = list(kwds.get("origin", [0.0] * self._dim))
        if len(origin) != self._dim:
            raise ValueError("origin must have %d dimensions" % self._dim)

        radian = angle * pi / 180.0
        cos_alpha, sin_alpha = cos(radian), sin(radian)

        for _idx, row in enumerate(self._coords):
            x, y = row[dim1] - origin[dim1], row[dim2] - origin[dim2]
            row[dim1] = cos_alpha * x - sin_alpha * y + origin[dim1]
            row[dim2] = sin_alpha * x + cos_alpha * y + origin[dim2]

    def scale(self, *args, **kwds):
        """Scales the layout.

        Scaling parameters can be provided either through the C{scale} keyword
        argument or through plain unnamed arguments. If a single integer or
        float is given, it is interpreted as a uniform multiplier to be applied
        on all dimensions. If it is a list or tuple, its length must be equal to
        the number of dimensions in the layout, and each element must be an
        integer or float describing the scaling coefficient in one of the
        dimensions.

        @keyword scale: scaling coefficients (integer, float, list or tuple)
        @keyword origin: the origin of scaling (this point will stay in place).
          Optional, defaults to the origin of the coordinate system being used.
        """
        origin = list(kwds.get("origin", [0.0] * self._dim))
        if len(origin) != self._dim:
            raise ValueError("origin must have %d dimensions" % self._dim)

        scaling = kwds.get("scale") or args
        if isinstance(scaling, (int, float)):
            scaling = [scaling]
        if len(scaling) == 0:
            raise ValueError("scaling factor must be given")
        elif len(scaling) == 1:
            if isinstance(scaling[0], (int, float)):
                scaling *= self._dim
            else:
                scaling = scaling[0]
        if len(scaling) != self._dim:
            raise ValueError("scaling factor list must have %d elements" % self._dim)

        for idx, row in enumerate(self._coords):
            self._coords[idx] = [
                (row[d] - origin[d]) * scaling[d] + origin[d] for d in range(self._dim)
            ]

    def translate(self, *args, **kwds):
        """Translates the layout.

        The translation vector can be provided either through the C{v} keyword
        argument or through plain unnamed arguments. If unnamed arguments are
        used, the vector can be supplied as a single list (or tuple) or just as
        a series of arguments. In all cases, the translation vector must have
        the same number of dimensions as the layout.

        @keyword v: the translation vector
        """
        v = kwds.get("v") or args
        if len(v) == 0:
            raise ValueError("translation vector must be given")
        elif len(v) == 1 and not isinstance(v[0], (int, float)):
            v = v[0]
        if len(v) != self._dim:
            raise ValueError("translation vector must have %d dimensions" % self._dim)

        for idx, row in enumerate(self._coords):
            self._coords[idx] = [row[d] + v[d] for d in range(self._dim)]

    def to_radial(self, min_angle=100, max_angle=80, min_radius=0.0, max_radius=1.0):
        """Converts a planar layout to a radial one

        This method applies only to 2D layouts. The X coordinate of the
        layout is transformed to an angle, with min(x) corresponding to
        the parameter called I{min_angle} and max(y) corresponding to
        I{max_angle}. Angles are given in degrees, zero degree corresponds
        to the direction pointing upwards. The Y coordinate is
        interpreted as a radius, with min(y) belonging to the minimum and
        max(y) to the maximum radius given in the arguments.

        This is not a fully generic polar coordinate transformation, but
        it is fairly useful in creating radial tree layouts from ordinary
        top-down ones (that's why the Y coordinate belongs to the radius).
        It can also be used in conjunction with the Fruchterman-Reingold
        layout algorithm via its I{miny} and I{maxy} parameters (see
        L{Graph.layout_fruchterman_reingold()<GraphBase.layout_fruchterman_reingold()>})
        to produce radial layouts where the radius belongs to some property of
        the vertices.

        @param min_angle: the angle corresponding to the minimum X value
        @param max_angle: the angle corresponding to the maximum X value
        @param min_radius: the radius corresponding to the minimum Y value
        @param max_radius: the radius corresponding to the maximum Y value
        """
        if self._dim != 2:
            raise TypeError("implemented only for 2D layouts")
        bbox = self.bounding_box()

        while min_angle > max_angle:
            max_angle += 360
        while min_angle > 360:
            min_angle -= 360
            max_angle -= 360
        while min_angle < 0:
            min_angle += 360
            max_angle += 360

        ratio_x = (max_angle - min_angle) / bbox.width
        ratio_x *= pi / 180.0
        min_angle *= pi / 180.0
        ratio_y = (max_radius - min_radius) / bbox.height
        for idx, (x, y) in enumerate(self._coords):
            alpha = (x - bbox.left) * ratio_x + min_angle
            radius = (y - bbox.top) * ratio_y + min_radius
            self._coords[idx] = [cos(alpha) * radius, -sin(alpha) * radius]

    def transform(self, function, *args, **kwds):
        """Performs an arbitrary transformation on the layout

        Additional positional and keyword arguments are passed intact to
        the given function.

        @param function: a function which receives the coordinates as a
          tuple and returns the transformed tuple.
        """
        self._coords = [
            list(function(tuple(row), *args, **kwds)) for row in self._coords
        ]

    def centroid(self):
        """Returns the centroid of the layout.

        The centroid of the layout is the arithmetic mean of the points in
        the layout.

        @return: the centroid as a list of floats"""
        centroid = [RunningMean() for _ in range(self._dim)]
        for row in self._coords:
            for dim in range(self._dim):
                centroid[dim].add(row[dim])
        return [rm.mean for rm in centroid]

    def boundaries(self, border=0):
        """Returns the boundaries of the layout.

        The boundaries are the minimum and maximum coordinates along all
        dimensions.

        @param border: this value gets subtracted from the minimum bounds
          and gets added to the maximum bounds before returning the coordinates
          of the box. Defaults to zero.
        @return: the minimum and maximum coordinates along all dimensions,
          in a tuple containing two lists, one for the minimum coordinates,
          the other one for the maximum.
        @raises ValueError: if the layout contains no layout items
        """
        if not self._coords:
            raise ValueError("layout contains no layout items")

        mins, maxs = [], []
        for dim in range(self._dim):
            col = [row[dim] for row in self._coords]
            mins.append(min(col) - border)
            maxs.append(max(col) + border)
        return mins, maxs

    def bounding_box(self, border=0):
        """Returns the bounding box of the layout.

        The bounding box of the layout is the smallest box enclosing all the
        points in the layout.

        @param border: this value gets subtracted from the minimum bounds
          and gets added to the maximum bounds before returning the coordinates
          of the box. Defaults to zero.
        @return: the coordinates of the lower left and the upper right corner
          of the box. "Lower left" means the minimum coordinates and "upper right"
          means the maximum. These are encapsulated in a L{BoundingBox} object.
        """
        if self._dim != 2:
            raise ValueError("Layout.boundary_box() supports 2D layouts only")

        try:
            (x0, y0), (x1, y1) = self.boundaries(border)
            return BoundingBox(x0, y0, x1, y1)
        except ValueError:
            return BoundingBox(0, 0, 0, 0)

    def center(self, *args, **kwds):
        """Centers the layout around the given point.

        The point itself can be supplied as multiple unnamed arguments, as a
        simple unnamed list or as a keyword argument. This operation moves
        the centroid of the layout to the given point. If no point is supplied,
        defaults to the origin of the coordinate system.

        @keyword p: the point where the centroid of the layout will be after
          the operation."""
        center = kwds.get("p") or args
        if len(center) == 0:
            center = [0.0] * self._dim
        elif len(center) == 1 and not isinstance(center[0], (int, float)):
            center = center[0]
        if len(center) != self._dim:
            raise ValueError("the given point must have %d dimensions" % self._dim)
        centroid = self.centroid()
        vec = [center[d] - centroid[d] for d in range(self._dim)]
        self.translate(vec)

    def copy(self):
        """Creates an exact copy of the layout."""
        return self.__copy__()

    def fit_into(self, bbox, keep_aspect_ratio=True):
        """Fits the layout into the given bounding box.

        The layout will be modified in-place.

        @param bbox: the bounding box in which to fit the layout. If the
          dimension of the layout is d, it can either be a d-tuple (defining
          the sizes of the box), a 2d-tuple (defining the coordinates of the
          top left and the bottom right point of the box), or a L{BoundingBox}
          object (for 2D layouts only).
        @param keep_aspect_ratio: whether to keep the aspect ratio of the current
          layout. If C{False}, the layout will be rescaled to fit exactly into
          the bounding box. If C{True}, the original aspect ratio of the layout
          will be kept and it will be centered within the bounding box.
        """
        if isinstance(bbox, BoundingBox):
            if self._dim != 2:
                raise TypeError("bounding boxes work for 2D layouts only")
            corner, target_sizes = [bbox.left, bbox.top], [bbox.width, bbox.height]
        elif len(bbox) == self._dim:
            corner, target_sizes = [0.0] * self._dim, list(bbox)
        elif len(bbox) == 2 * self._dim:
            corner, opposite_corner = list(bbox[0 : self._dim]), list(bbox[self._dim :])
            for i in range(self._dim):
                if corner[i] > opposite_corner[i]:
                    corner[i], opposite_corner[i] = opposite_corner[i], corner[i]
            target_sizes = [
                max_val - min_val for min_val, max_val in zip(corner, opposite_corner)
            ]

        try:
            mins, maxs = self.boundaries()
        except ValueError:
            mins, maxs = [0.0] * self._dim, [0.0] * self._dim
        sizes = [max_val - min_val for min_val, max_val in zip(mins, maxs)]

        for i, size in enumerate(sizes):
            if size == 0:
                sizes[i] = 2
                mins[i] -= 1
                maxs[i] += 1

        ratios = [
            float(target_size) / current_size
            for current_size, target_size in zip(sizes, target_sizes)
        ]
        if keep_aspect_ratio:
            min_ratio = min(ratios)
            ratios = [min_ratio] * self._dim

        translations = []
        for i in range(self._dim):
            trans = (target_sizes[i] - ratios[i] * sizes[i]) / 2.0
            trans -= mins[i] * ratios[i] - corner[i]
            translations.append(trans)

        self.scale(*ratios)
        self.translate(*translations)


def _layout(graph, layout=None, *args, **kwds):
    """Returns the layout of the graph according to a layout algorithm.

    Parameters and keyword arguments not specified here are passed to the
    layout algorithm directly. See the documentation of the layout
    algorithms for the explanation of these parameters.

    Registered layout names understood by this method are:

      - C{auto}, C{automatic}: automatic layout
        (see L{Graph.layout_auto})

      - C{bipartite}: bipartite layout (see L{GraphBase.layout_bipartite})

      - C{circle}, C{circular}: circular layout
        (see L{GraphBase.layout_circle})

      - C{dh}, C{davidson_harel}: Davidson-Harel layout (see
        L{GraphBase.layout_davidson_harel})

      - C{drl}: DrL layout for large graphs (see L{GraphBase.layout_drl})

      - C{drl_3d}: 3D DrL layout for large graphs
        (see L{GraphBase.layout_drl})

      - C{fr}, C{fruchterman_reingold}: Fruchterman-Reingold layout
        (see L{GraphBase.layout_fruchterman_reingold}).

      - C{fr_3d}, C{fr3d}, C{fruchterman_reingold_3d}: 3D Fruchterman-
        Reingold layout (see L{GraphBase.layout_fruchterman_reingold}).

      - C{grid}: regular grid layout in 2D (see L{GraphBase.layout_grid})

      - C{grid_3d}: regular grid layout in 3D (see L{GraphBase.layout_grid})

      - C{graphopt}: the graphopt algorithm (see L{GraphBase.layout_graphopt})

      - C{kk}, C{kamada_kawai}: Kamada-Kawai layout
        (see L{GraphBase.layout_kamada_kawai})

      - C{kk_3d}, C{kk3d}, C{kamada_kawai_3d}: 3D Kamada-Kawai layout
        (see L{GraphBase.layout_kamada_kawai})

      - C{lgl}, C{large}, C{large_graph}: Large Graph Layout
        (see L{GraphBase.layout_lgl})

      - C{mds}: multidimensional scaling layout (see L{GraphBase.layout_mds})

      - C{random}: random layout (see L{GraphBase.layout_random})

      - C{random_3d}: random 3D layout (see L{GraphBase.layout_random})

      - C{rt}, C{tree}, C{reingold_tilford}: Reingold-Tilford tree
        layout (see L{GraphBase.layout_reingold_tilford})

      - C{rt_circular}, C{reingold_tilford_circular}: circular
        Reingold-Tilford tree layout
        (see L{GraphBase.layout_reingold_tilford_circular})

      - C{sphere}, C{spherical}, C{circle_3d}, C{circular_3d}: spherical
        layout (see L{GraphBase.layout_circle})

      - C{star}: star layout (see L{GraphBase.layout_star})

      - C{sugiyama}: Sugiyama layout (see L{Graph.layout_sugiyama})

    @param layout: the layout to use. This can be one of the registered
      layout names or a callable which returns either a L{Layout} object or
      a list of lists containing the coordinates. If C{None}, uses the
      value of the C{plotting.layout} configuration key.
    @return: a L{Layout} object.
    """
    # Deferred import to avoid cycles
    from igraph import config

    if layout is None:
        layout = config["plotting.layout"]
    if callable(layout):
        method = layout
    else:
        layout = layout.lower()
        if layout[-3:] == "_3d":
            kwds["dim"] = 3
            layout = layout[:-3]
        elif layout[-2:] == "3d":
            kwds["dim"] = 3
            layout = layout[:-2]
        method = getattr(graph.__class__, graph._layout_mapping[layout])
    if not callable(method):
        raise ValueError("layout method must be callable")
    layout = method(graph, *args, **kwds)
    if not isinstance(layout, Layout):
        layout = Layout(layout)
    return layout


def align_layout(graph, layout):
    """Aligns a graph layout with the coordinate axes

    This function centers a vertex layout on the coordinate system origin and
    rotates the layout to achieve a visually pleasing alignment with the coordinate
    axes. Doing this is particularly useful with force-directed layouts such as
    L{Graph.layout_fruchterman_reingold}. Layouts in arbitrary dimensional spaces
    are supported.

    @param graph: the graph that the layout is associated with.
    @param layout: the L{Layout} object containing the vertex coordinates
      to align.
    @return: a new aligned L{Layout} object.
    """
    from igraph._igraph import _align_layout

    if not isinstance(layout, Layout):
        layout = Layout(layout)

    return Layout(_align_layout(graph, layout.coords))


def _layout_auto(graph, *args, **kwds):
    """Chooses and runs a suitable layout function based on simple
    topological properties of the graph.

    This function tries to choose an appropriate layout function for
    the graph using the following rules:

      1. If the graph has an attribute called C{layout}, it will be
         used. It may either be a L{Layout} instance, a list of
         coordinate pairs, the name of a layout function, or a
         callable function which generates the layout when called
         with the graph as a parameter.

      2. Otherwise, if the graph has vertex attributes called C{x}
         and C{y}, these will be used as coordinates in the layout.
         When a 3D layout is requested (by setting C{dim} to 3),
         a vertex attribute named C{z} will also be needed.

      3. Otherwise, if the graph is connected and has at most 100
         vertices, the Kamada-Kawai layout will be used (see
         L{GraphBase.layout_kamada_kawai()}).

      4. Otherwise, if the graph has at most 1000 vertices, the
         Fruchterman-Reingold layout will be used (see
         L{GraphBase.layout_fruchterman_reingold()}).

      5. If everything else above failed, the DrL layout algorithm
         will be used (see L{GraphBase.layout_drl()}).

    All the arguments of this function except C{dim} are passed on
    to the chosen layout function (in case we have to call some layout
    function).

    @keyword dim: specifies whether we would like to obtain a 2D or a
      3D layout.
    @return: a L{Layout} object.
    """
    if "layout" in graph.attributes():
        layout = graph["layout"]
        if isinstance(layout, Layout):
            # Layouts are used intact
            return layout
        if isinstance(layout, (list, tuple)):
            # Lists/tuples are converted to layouts
            return Layout(layout)
        if callable(layout):
            # Callables are called
            return Layout(layout(*args, **kwds))
        # Try Graph.layout()
        return graph.layout(layout, *args, **kwds)

    dim = kwds.get("dim", 2)
    vattrs = graph.vertex_attributes()
    if "x" in vattrs and "y" in vattrs:
        if dim == 3 and "z" in vattrs:
            return Layout(list(zip(graph.vs["x"], graph.vs["y"], graph.vs["z"])))
        else:
            return Layout(list(zip(graph.vs["x"], graph.vs["y"])))

    if graph.vcount() <= 100 and graph.is_connected():
        algo = "kk"
    elif graph.vcount() <= 1000:
        algo = "fr"
    else:
        algo = "drl"
    return graph.layout(algo, *args, **kwds)


def _layout_sugiyama(
    graph,
    layers=None,
    weights=None,
    hgap=1,
    vgap=1,
    maxiter=100,
):
    """Places the vertices using a layered Sugiyama layout.

    This is a layered layout that is most suitable for directed acyclic graphs,
    although it works on undirected or cyclic graphs as well.

    Each vertex is assigned to a layer and each layer is placed on a horizontal
    line. Vertices within the same layer are then permuted using the barycenter
    heuristic that tries to minimize edge crossings.

    Dummy vertices will be added on edges that span more than one layer. The
    returned layout therefore contains more rows than the number of nodes in
    the original graph; the extra rows correspond to the dummy vertices.

    B{References}:

      - K Sugiyama, S Tagawa, M Toda: Methods for visual understanding of
        hierarchical system structures. I{IEEE Systems, Man and Cybernetics}
        11(2):109-125, 1981.

      - P Eades, X Lin and WF Smyth: A fast effective heuristic for the
        feedback arc set problem. I{Information Processing Letters} 47:319-323, 1993.

    @param layers: a vector specifying a non-negative integer layer index for
      each vertex, or the name of a numeric vertex attribute that contains
      the layer indices. If C{None}, a layering will be determined
      automatically. For undirected graphs, a spanning tree will be extracted
      and vertices will be assigned to layers using a breadth first search from
      the node with the largest degree. For directed graphs, cycles are broken
      by reversing the direction of edges in an approximate feedback arc set
      using the heuristic of Eades, Lin and Smyth, and then using longest path
      layering to place the vertices in layers.
    @param weights: edge weights to be used. Can be a sequence or iterable or
      even an edge attribute name.
    @param hgap: minimum horizontal gap between vertices in the same layer.
    @param vgap: vertical gap between layers. The layer index will be
      multiplied by I{vgap} to obtain the Y coordinate.
    @param maxiter: maximum number of iterations to take in the crossing
      reduction step. Increase this if you feel that you are getting too many
      edge crossings.
    @return: the calculated layout and an additional list of matrices where the
      i-th matrix contains the control points of edge I{i} in the original graph
      (or an empty matrix if no control points are needed on the edge)
    """
    coords, routing = GraphBase._layout_sugiyama(
        graph, layers, weights, hgap, vgap, maxiter
    )
    layout = Layout(coords)
    layout.edge_routing = routing
    return layout


def _layout_method_wrapper(func):
    """Wraps an existing layout method to ensure that it returns a Layout
    instead of a list of lists.

    @param func: the method to wrap. Must be a method of the Graph object.
    @return: a new method
    """

    def result(*args, **kwds):
        layout = func(*args, **kwds)
        if not isinstance(layout, Layout):
            layout = Layout(layout)
        return layout

    result.__name__ = func.__name__
    result.__doc__ = func.__doc__
    return result


def _3d_version_for(func):
    """Creates an alias for the 3D version of the given layout algoritm.

    This function is a decorator that creates a method which calls I{func} after
    attaching C{dim=3} to the list of keyword arguments.

    @param func: must be a method of the Graph object.
    @return: a new method
    """

    def result(*args, **kwds):
        kwds["dim"] = 3
        return func(*args, **kwds)

    result.__name__ = "%s_3d" % func.__name__
    result.__doc__ = """Alias for L{%s()} with dim=3.\n\n@see: Graph.%s()""" % (
        func.__name__,
        func.__name__,
    )
    return result


# After adjusting something here, don't forget to update the docstring
# of Graph.layout if necessary!
_layout_mapping = {
    "auto": "layout_auto",
    "automatic": "layout_auto",
    "bipartite": "layout_bipartite",
    "circle": "layout_circle",
    "circular": "layout_circle",
    "davidson_harel": "layout_davidson_harel",
    "dh": "layout_davidson_harel",
    "drl": "layout_drl",
    "fr": "layout_fruchterman_reingold",
    "fruchterman_reingold": "layout_fruchterman_reingold",
    "graphopt": "layout_graphopt",
    "grid": "layout_grid",
    "kk": "layout_kamada_kawai",
    "kamada_kawai": "layout_kamada_kawai",
    "lgl": "layout_lgl",
    "large": "layout_lgl",
    "large_graph": "layout_lgl",
    "mds": "layout_mds",
    "random": "layout_random",
    "rt": "layout_reingold_tilford",
    "tree": "layout_reingold_tilford",
    "reingold_tilford": "layout_reingold_tilford",
    "rt_circular": "layout_reingold_tilford_circular",
    "reingold_tilford_circular": "layout_reingold_tilford_circular",
    "sphere": "layout_sphere",
    "spherical": "layout_sphere",
    "star": "layout_star",
    "sugiyama": "layout_sugiyama",
}



"""

.. _tutorials-dag:

======================
Directed Acyclic Graph
======================

This example demonstrates how to create a random directed acyclic graph (DAG), which is useful in a number of contexts including for Git commit history.
"""

import igraph as ig
import matplotlib.pyplot as plt
import random


# %%
# First, we set a random seed for reproducibility.
random.seed(0)

# %%
# First, we generate a random undirected graph with a fixed number of edges, without loops.
g = ig.Graph.Erdos_Renyi(n=15, m=30, directed=False, loops=False)

# %%
# Then we convert it to a DAG *in place*. This method samples DAGs with a given number of edges and vertices uniformly.
g.to_directed(mode="acyclic")

# %%
# We can print out a summary of the DAG.
ig.summary(g)


# %%
# Finally, we can plot the graph using the Sugiyama layout from :meth:`igraph.Graph.layout_sugiyama`:
fig, ax = plt.subplots()
ig.plot(
    g,
    target=ax,
    layout="sugiyama",
    vertex_size=15,
    vertex_color="grey",
    edge_color="#222",
    edge_width=1,
)
plt.show()



import unittest

from igraph import Graph


class MotifTests(unittest.TestCase):
    def setUp(self):
        self.g = Graph.Erdos_Renyi(100, 0.2, directed=True)

    def testDyads(self):
        # @note: this test is not exhaustive, it only checks whether the
        # L{DyadCensus} objects "understand" attribute and item accessors
        dc = self.g.dyad_census()
        accessors = ["mut", "mutual", "asym", "asymm", "asymmetric", "null"]
        for a in accessors:
            self.assertTrue(isinstance(getattr(dc, a), int))
            self.assertTrue(isinstance(dc[a], int))
        self.assertTrue(isinstance(list(dc), list))
        self.assertTrue(isinstance(tuple(dc), tuple))
        self.assertTrue(len(list(dc)) == 3)
        self.assertTrue(len(tuple(dc)) == 3)

    def testTriads(self):
        # @note: this test is not exhaustive, it only checks whether the
        # L{TriadCensus} objects "understand" attribute and item accessors
        tc = self.g.triad_census()
        accessors = ["003", "012", "021d", "030C"]
        for a in accessors:
            self.assertTrue(isinstance(getattr(tc, "t" + a), int))
            self.assertTrue(isinstance(tc[a], int))
        self.assertTrue(isinstance(list(tc), list))
        self.assertTrue(isinstance(tuple(tc), tuple))
        self.assertTrue(len(list(tc)) == 16)
        self.assertTrue(len(tuple(tc)) == 16)


class TrianglesTests(unittest.TestCase):
    def testListTriangles(self):
        g = Graph.Famous("petersen")
        self.assertEqual([], g.list_triangles())

        g = Graph([(0, 1), (1, 2), (2, 0), (1, 3), (3, 2), (4, 2), (4, 3)])
        observed = g.list_triangles()
        self.assertTrue(all(isinstance(x, tuple) for x in observed))
        observed = sorted(sorted(tri) for tri in observed)
        expected = [[0, 1, 2], [1, 2, 3], [2, 3, 4]]
        self.assertEqual(observed, expected)

        g = Graph.GRG(100, 0.2)
        tri = Graph.Full(3)
        observed = sorted(sorted(tri) for tri in g.list_triangles())
        expected = sorted(x for x in g.get_subisomorphisms_vf2(tri) if x == sorted(x))
        self.assertEqual(observed, expected)


def suite():
    motif_suite = unittest.defaultTestLoader.loadTestsFromTestCase(MotifTests)
    triangles_suite = unittest.defaultTestLoader.loadTestsFromTestCase(TrianglesTests)
    return unittest.TestSuite([motif_suite, triangles_suite])


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()



from igraph.drawing.utils import FakeModule
from typing import Any

__all__ = ("find_matplotlib",)
__docformat__ = "restructuredtext en"


def find_matplotlib() -> Any:
    """Tries to import the ``matplotlib`` Python module if it is installed.
    Returns a fake module if everything fails.
    """
    try:
        import matplotlib as mpl

        has_mpl = True
    except ImportError:
        mpl = FakeModule("You need to install matplotlib to use this functionality")
        has_mpl = False

    if has_mpl:
        import matplotlib.pyplot as plt
    else:
        plt = FakeModule(
            "You need to install matplotlib.pyplot to use this functionality"
        )

    return mpl, plt



# vim:ts=4:sw=4:sts=4:et
# -*- coding: utf-8 -*-
"""Implementation of Python-level sparse matrix operations."""

from __future__ import with_statement

__all__ = ()
__docformat__ = "restructuredtext en"

from operator import add
from igraph._igraph import (
    ADJ_DIRECTED,
    ADJ_UNDIRECTED,
    ADJ_MAX,
    ADJ_MIN,
    ADJ_PLUS,
    ADJ_UPPER,
    ADJ_LOWER,
)


_SUPPORTED_MODES = ("directed", "undirected", "max", "min", "plus", "lower", "upper")


def _convert_mode_argument(mode):
    # resolve mode constants, convert to lowercase
    mode = {
        ADJ_DIRECTED: "directed",
        ADJ_UNDIRECTED: "undirected",
        ADJ_MAX: "max",
        ADJ_MIN: "min",
        ADJ_PLUS: "plus",
        ADJ_UPPER: "upper",
        ADJ_LOWER: "lower",
    }.get(mode, mode).lower()

    if mode not in _SUPPORTED_MODES:
        raise ValueError("mode should be one of " + (" ".join(_SUPPORTED_MODES)))

    if mode == "undirected":
        mode = "max"

    return mode


def _maybe_halve_diagonal(m, condition):
    """Halves all items in the diagonal of the given SciPy sparse matrix in
    coo mode *if* and *only if* the given condition is True.

    Returns the row, column and data arrays.
    """
    data_array = m.data
    if condition:
        # We can't do data_array[m.row == m.col] /= 2 here because we would be
        # modifying the array in-place and the end user wouldn't like if we
        # messed with their matrix. So we make a copy.
        data_array = data_array.copy()
        (idxs,) = (m.row == m.col).nonzero()
        for i in idxs:
            data_array[i] /= 2

    return m.row, m.col, data_array


# Logic to get graph from scipy sparse matrix. This would be simple if there
# weren't so many modes.
def _graph_from_sparse_matrix(klass, matrix, mode="directed", loops="once"):
    """Construct graph from sparse matrix, unweighted.

    @param loops: specifies how the diagonal of the matrix should be handled:

      - C{"ignore"} - ignore loop edges in the diagonal
      - C{"once"} - treat the diagonal entries as loop edge counts
      - C{"twice"} - treat the diagonal entries as I{twice} the number
        of loop edges
    """
    # This function assumes there is scipy and the matrix is a scipy sparse
    # matrix. The caller should make sure those conditions are met.
    from scipy import sparse

    if not isinstance(matrix, sparse.coo_matrix):
        matrix = matrix.tocoo()

    nvert = max(matrix.shape)
    if min(matrix.shape) != nvert:
        raise ValueError("Matrix must be square")

    # Shorthand notation
    m = matrix

    mode = _convert_mode_argument(mode)

    keep_loops = loops == "twice" or loops == "once" or loops is True
    m_row, m_col, m_data = _maybe_halve_diagonal(m, loops == "twice")

    if mode == "directed":
        edges = []
        for i, j, n in zip(m_row, m_col, m_data):
            if i != j or keep_loops:
                edges.extend([(i, j)] * n)

    elif mode in ("max", "plus"):
        fun = max if mode == "max" else add
        nedges = {}
        for i, j, n in zip(m_row, m_col, m_data):
            if i == j and not keep_loops:
                continue
            pair = (i, j) if i < j else (j, i)
            nedges[pair] = fun(nedges.get(pair, 0), n)

        edges = sum(
            ([e] * n for e, n in nedges.items()),
            [],
        )

    elif mode == "min":
        tmp = {(i, j): n for i, j, n in zip(m_row, m_col, m_data)}

        nedges = {}
        for pair, weight in tmp.items():
            i, j = pair
            if i == j and keep_loops:
                nedges[pair] = weight
            elif i < j:
                nedges[pair] = min(weight, tmp.get((j, i), 0))

        edges = sum(
            ([e] * n for e, n in nedges.items()),
            [],
        )

    elif mode == "upper":
        edges = []
        for i, j, n in zip(m_row, m_col, m_data):
            if j > i or (keep_loops and j == i):
                edges.extend([(i, j)] * n)

    elif mode == "lower":
        edges = []
        for i, j, n in zip(m_row, m_col, m_data):
            if j < i or (keep_loops and j == i):
                edges.extend([(i, j)] * n)

    else:
        raise ValueError(f"invalid mode: {mode!r}")

    return klass(nvert, edges=edges, directed=(mode == "directed"))


def _graph_from_weighted_sparse_matrix(
    klass, matrix, mode=ADJ_DIRECTED, attr="weight", loops="once"
):
    """Construct graph from sparse matrix, weighted

    NOTE: Of course, you cannot emcompass a fully general weighted multigraph
    with a single adjacency matrix, so we don't try to do it here either.

    @param loops: specifies how to handle loop edges. When C{False} or
        C{"ignore"}, the diagonal of the adjacency matrix will be ignored. When
        C{True} or C{"once"}, the diagonal is assumed to contain the weight of the
        corresponding loop edge. When C{"twice"}, the diagonal is assumed to
        contain I{twice} the weight of the corresponding loop edge.
    """
    # This function assumes there is scipy and the matrix is a scipy sparse
    # matrix. The caller should make sure those conditions are met.
    from scipy import sparse

    if not isinstance(matrix, sparse.coo_matrix):
        matrix = matrix.tocoo()

    nvert = max(matrix.shape)
    if min(matrix.shape) != nvert:
        raise ValueError("Matrix must be square")

    # Shorthand notation
    m = matrix

    mode = _convert_mode_argument(mode)

    keep_loops = loops == "twice" or loops == "once" or loops is True
    m_row, m_col, m_data = _maybe_halve_diagonal(m, loops == "twice")

    if mode == "directed":
        if not keep_loops:
            edges, weights = [], []
            for i, j, n in zip(m_row, m_col, m_data):
                if i != j:
                    edges.append((i, j))
                    weights.append(n)
        else:  # loops == "once" or True
            edges = list(zip(m_row, m_col))
            weights = list(m_data)

    elif mode in ("max", "plus"):
        fun = max if mode == "max" else add
        nedges = {}
        for i, j, n in zip(m_row, m_col, m_data):
            if i == j and not keep_loops:
                continue

            pair = (i, j) if i < j else (j, i)
            nedges[pair] = fun(nedges.get(pair, 0), n)

        edges, weights = zip(*nedges.items())

    elif mode == "min":
        tmp = {(i, j): n for i, j, n in zip(m_row, m_col, m_data)}

        nedges = {}
        for pair, weight in tmp.items():
            i, j = pair
            if i == j and keep_loops:
                nedges[pair] = weight
            elif i < j:
                nedges[pair] = min(weight, tmp.get((j, i), 0))

        edges, weights = [], []
        for pair in sorted(nedges.keys()):
            weight = nedges[pair]
            if weight != 0:
                edges.append(pair)
                weights.append(nedges[pair])

    elif mode == "upper":
        edges, weights = [], []
        for i, j, n in zip(m_row, m_col, m_data):
            if j > i or (keep_loops and j == i):
                edges.append((i, j))
                weights.append(n)

    elif mode == "lower":
        edges, weights = [], []
        for i, j, n in zip(m_row, m_col, m_data):
            if j < i or (keep_loops and j == i):
                edges.append((i, j))
                weights.append(n)

    else:
        raise ValueError(f"invalid mode: {mode!r}")

    return klass(
        nvert, edges=edges, directed=(mode == "directed"), edge_attrs={attr: weights}
    )



"""Drawers for various edge styles in Matplotlib graph plots."""

from math import atan2, cos, pi, sin

from igraph.drawing.baseclasses import AbstractEdgeDrawer
from igraph.drawing.metamagic import AttributeCollectorBase
from igraph.drawing.matplotlib.utils import find_matplotlib
from igraph.drawing.utils import (
    get_bezier_control_points_for_curved_edge,
    FakeModule,
)

__all__ = ("MatplotlibEdgeDrawer", "EdgeCollection")

mpl, plt = find_matplotlib()
try:
    PatchCollection = mpl.collections.PatchCollection
except AttributeError:
    PatchCollection = FakeModule


class MatplotlibEdgeDrawer(AbstractEdgeDrawer):
    """Matplotlib-specific abstract edge drawer object."""

    def __init__(self, context, palette):
        """Constructs the edge drawer.

        @param context: a Matplotlib axes object on which the edges will be
            drawn.
        @param palette: the palette that can be used to map integer color
            indices to colors when drawing edges
        """
        self.context = context
        self.palette = palette
        self.VisualEdgeBuilder = self._construct_visual_edge_builder()

    def _construct_visual_edge_builder(self):
        """Construct the visual edge builder that will collect the visual
        attributes of an edge when it is being drawn."""

        class VisualEdgeBuilder(AttributeCollectorBase):
            """Builder that collects some visual properties of an edge for
            drawing"""

            _kwds_prefix = "edge_"
            arrow_size = 15
            arrow_width = 15
            color = ("#444", self.palette.get)
            curved = (0.0, self._curvature_to_float)
            label = None
            label_color = ("black", self.palette.get)
            label_size = 12.0
            font = "sans-serif"
            width = 2.0
            background = None
            align_label = False
            zorder = 1
            loop_size = 30

        return VisualEdgeBuilder

    def build_patch(self, edge, src_vertex, dest_vertex):
        art = mpl.patches.PathPatch(
            mpl.path.Path([[0, 0]]),
            edgecolor=edge.color,
            facecolor=edge.color if src_vertex != dest_vertex else "none",
            linewidth=edge.width,
            zorder=edge.zorder,
            clip_on=True,
        )
        return art

    # The following two methods are replaced by dummy functions, the rest is
    # taken care of in EdgeCollection for efficiency
    def draw_directed_edge(self, edge, src_vertex, dest_vertex):
        return self.build_patch(edge)

    def draw_undirected_edge(self, edge, src_vertex, dest_vertex):
        return self.build_patch(edge)


class EdgeCollection(PatchCollection):
    def __init__(self, *args, **kwargs):
        kwargs["match_original"] = True
        self._visual_vertices = kwargs.pop("visual_vertices", None)
        self._directed = kwargs.pop("directed", False)
        self._arrow_sizes = kwargs.pop("arrow_sizes", None)
        self._arrow_widths = kwargs.pop("arrow_widths", None)
        self._loop_sizes = kwargs.pop("loop_sizes", None)
        self._curved = kwargs.pop("curved", None)
        super().__init__(*args, **kwargs)

    @staticmethod
    def _get_edge_vertex_sizes(edge_vertices):
        sizes = []
        for visual_vertex in edge_vertices:
            if visual_vertex.size is not None:
                sizes.append(visual_vertex.size)
            else:
                sizes.append(max(visual_vertex.width, visual_vertex.height))
        return sizes

    @staticmethod
    def _compute_edge_angles(path, trans, directed, curved):
        """Compute edge angles for both starting and ending vertices.

        NOTE: The domain of atan2 is (-pi, pi].
        """
        positions = trans(path.vertices)

        # first angle
        if not directed:
            x1, y1 = positions[0]
            x2, y2 = positions[1]
        elif not curved:
            x1, y1 = positions[1]
            x2, y2 = positions[0]
        else:
            x1, y1 = positions[3]
            x2, y2 = positions[2]
        angle1 = atan2(y2 - y1, x2 - x1)

        # second angle
        if not directed:
            x1, y1 = positions[-1]
            x2, y2 = positions[-2]
        else:
            x1, y1 = positions[-3]
            x2, y2 = positions[-1]
        angle2 = atan2(y2 - y1, x2 - x1)
        return (angle1, angle2)

    def _compute_paths(self, transform=None):
        import numpy as np

        visual_vertices = self._visual_vertices
        if transform is None:
            transform = self.get_transform()
        trans = transform.transform
        trans_inv = transform.inverted().transform

        # Loops split the largest wedge left open by other
        # edges of that vertex. The algo is:
        # (i) Find what vertices each loop belongs to
        # (ii) While going through the edges, record the angles
        #      for vertices with loops
        # (iii) Plot each loop based on the recorded angles
        loop_vertex_dict = {}
        for i, edge_vertices in enumerate(visual_vertices):
            if edge_vertices[0] == edge_vertices[1]:
                if edge_vertices[0] not in loop_vertex_dict:
                    loop_vertex_dict[edge_vertices[0]] = {
                        "indices": [],
                        "sizes": [],
                        "edge_angles": [],
                    }
                    if self._directed:
                        loop_vertex_dict[edge_vertices[0]]["arrow_sizes"] = []
                        loop_vertex_dict[edge_vertices[0]]["arrow_widths"] = []
                loop_vertex_dict[edge_vertices[0]]["indices"].append(i)

        # Get actual coordinates of the vertex border (rough)
        paths = []
        for i, edge_vertices in enumerate(visual_vertices):
            if self._directed:
                if (self._arrow_sizes is None) or (self._arrow_widths is None):
                    arrow_size = 0
                    arrow_width = 0
                else:
                    arrow_size = self._arrow_sizes[i]
                    arrow_width = self._arrow_widths[i]

            # Loops are positioned post-facto in the space left by the othter edges
            if edge_vertices[0] == edge_vertices[1]:
                paths.append(None)
                loop_vertex_dict[edge_vertices[0]]["sizes"].append(
                    self._loop_sizes[i],
                )
                if self._directed:
                    loop_vertex_dict[edge_vertices[0]]["arrow_sizes"].append(
                        arrow_size,
                    )
                    loop_vertex_dict[edge_vertices[0]]["arrow_widths"].append(
                        arrow_width,
                    )
                continue

            coords = np.vstack(
                [
                    edge_vertices[0].position,
                    edge_vertices[1].position,
                ]
            )
            coordst = trans(coords)
            sizes = self._get_edge_vertex_sizes(edge_vertices)
            if self._curved is not None:
                curved = self._curved[i]
            else:
                curved = False

            if self._directed:
                path = self._compute_path_directed(
                    coordst,
                    sizes,
                    trans_inv,
                    curved,
                    arrow_size,
                    arrow_width,
                )
            else:
                path = self._compute_path_undirected(
                    coordst,
                    sizes,
                    trans_inv,
                    curved,
                )

            # Collect angles for this vertex, to be used for loops plotting below
            angles = self._compute_edge_angles(path, trans, self._directed, curved)
            if edge_vertices[0] in loop_vertex_dict:
                loop_vertex_dict[edge_vertices[0]]["edge_angles"].append(angles[0])
            if edge_vertices[1] in loop_vertex_dict:
                loop_vertex_dict[edge_vertices[1]]["edge_angles"].append(angles[1])

            # Add the path for this non-loop edge
            paths.append(path)

        # Deal with loops at the end
        for visual_vertex, ldict in loop_vertex_dict.items():
            coords = np.vstack([visual_vertex.position] * 2)
            coordst = trans(coords)
            vertex_size = self._get_edge_vertex_sizes([visual_vertex])[0]

            edge_angles = ldict["edge_angles"]
            if edge_angles:
                edge_angles.sort()
                # Circle around
                edge_angles.append(edge_angles[0] + 2 * pi)
                wedges = [
                    (a2 - a1) for a1, a2 in zip(edge_angles[:-1], edge_angles[1:])
                ]
                # Argsort
                imax = max(range(len(wedges)), key=lambda i: wedges[i])
                angle1, angle2 = edge_angles[imax], edge_angles[imax + 1]
            else:
                # Isolated vertices with loops
                angle1, angle2 = -pi, pi

            nloops = len(ldict["indices"])
            for i in range(nloops):
                angle1i = angle1 + (angle2 - angle1) * i / nloops
                angle2i = angle1 + (angle2 - angle1) * (i + 1) / nloops
                if self._directed:
                    loop_kwargs = {
                        "arrow_size": ldict["arrow_sizes"][i],
                        "arrow_width": ldict["arrow_widths"][i],
                    }
                else:
                    loop_kwargs = {}
                path = self._compute_path_loop(
                    coordst[0],
                    vertex_size,
                    ldict["sizes"][i],
                    angle1i,
                    angle2i,
                    trans_inv,
                    angle_padding_fraction=0.1,
                    **loop_kwargs,
                )
                paths[ldict["indices"][i]] = path

        return paths

    def _compute_path_loop(
        self,
        coordt,
        vertex_size,
        loop_size,
        angle1,
        angle2,
        trans_inv,
        angle_padding_fraction=0.1,
        arrow_size=None,
        arrow_width=None,
    ):
        import numpy as np

        # Special argument for loop size to scale with vertices
        if loop_size < 0:
            loop_size = -loop_size * vertex_size

        # Pad angles to make a little space for tight arrowheads
        angle1, angle2 = (
            angle1 * (1 - angle_padding_fraction) + angle2 * angle_padding_fraction,
            angle1 * angle_padding_fraction + angle2 * (1 - angle_padding_fraction),
        )

        # Too large wedges, just use a quarter
        if angle2 - angle1 > pi / 3:
            angle_mid = (angle2 + angle1) * 0.5
            angle1 = angle_mid - pi / 6
            angle2 = angle_mid + pi / 6

        start = vertex_size / 2 * np.array([cos(angle1), sin(angle1)])
        end = vertex_size / 2 * np.array([cos(angle2), sin(angle2)])
        amix = 0.05
        aux1 = loop_size * np.array(
            [
                cos(angle1 * (1 - amix) + angle2 * amix),
                sin(angle1 * (1 - amix) + angle2 * amix),
            ]
        )
        aux2 = loop_size * np.array(
            [
                cos(angle1 * amix + angle2 * (1 - amix)),
                sin(angle1 * amix + angle2 * (1 - amix)),
            ]
        )

        if not self._directed:
            vertices = np.vstack(
                [
                    start,
                    aux1,
                    aux2,
                    end,
                    aux2,
                    aux1,
                    start,
                ]
            )
            codes = ["MOVETO"] + ["CURVE4"] * 6
        else:
            # Tweak the bezier points
            aux1 *= (loop_size + arrow_size) / loop_size
            aux2 *= (loop_size + arrow_size) / loop_size
            # Angle between end/tip and vertex centre
            theta = angle2
            voff_unity = np.array([cos(theta), sin(theta)])
            voff_unity_90 = voff_unity @ [[0, 1], [-1, 0]]
            headbase = end + arrow_size * voff_unity
            headleft = headbase + 0.5 * arrow_width * voff_unity_90
            headright = headbase - 0.5 * arrow_width * voff_unity_90
            vertices = np.vstack(
                [
                    start,
                    aux1,
                    aux2,
                    headbase,
                    headleft,
                    end,
                    headright,
                    headbase,
                    aux2,
                    aux1,
                    start,
                ]
            )
            codes = ["MOVETO"] + ["CURVE4"] * 3 + ["LINETO"] * 4 + ["CURVE4"] * 3

        # Offset to place and transform to data coordinates
        vertices = trans_inv(coordt + vertices)
        codes = [getattr(mpl.path.Path, x) for x in codes]
        path = mpl.path.Path(
            vertices,
            codes=codes,
        )
        return path

    def _compute_path_undirected(self, coordst, sizes, trans_inv, curved):
        path = {"vertices": [], "codes": []}
        path["codes"].append("MOVETO")
        if not curved:
            path["codes"].append("LINETO")

            # Start
            theta = atan2(*((coordst[1] - coordst[0])[::-1]))
            voff = 0 * coordst[0]
            voff[:] = [cos(theta), sin(theta)]
            voff *= sizes[0] / 2
            path["vertices"].append(coordst[0] + voff)

            # End
            voff[:] = [cos(theta), sin(theta)]
            voff *= sizes[1] / 2
            path["vertices"].append(coordst[1] - voff)
        else:
            path["codes"].extend(["CURVE4"] * 3)

            aux1, aux2 = get_bezier_control_points_for_curved_edge(
                *coordst.ravel(),
                curved,
            )

            # Start
            theta = atan2(*((aux1 - coordst[0])[::-1]))
            voff = 0 * coordst[0]
            voff[:] = [cos(theta), sin(theta)]
            voff *= sizes[0] / 2
            path["vertices"].append(coordst[0] + voff)

            # Bezier
            path["vertices"].append(aux1)
            path["vertices"].append(aux2)

            # End
            theta = atan2(*((coordst[1] - aux2)[::-1]))
            voff = 0 * coordst[0]
            voff[:] = [cos(theta), sin(theta)]
            voff *= sizes[1] / 2
            path["vertices"].append(coordst[1] - voff)

            # This is a dirty trick to make the facecolor work
            # without making a separate Patch, which would be a little messy
            path["codes"].extend(["CURVE4"] * 3)
            path["vertices"].extend(path["vertices"][-2::-1])

        path = mpl.path.Path(
            path["vertices"],
            codes=[getattr(mpl.path.Path, x) for x in path["codes"]],
        )
        path.vertices = trans_inv(path.vertices)
        return path

    def _compute_path_directed(
        self, coordst, sizes, trans_inv, curved, arrow_size, arrow_width
    ):
        path = {"vertices": [], "codes": []}
        path["codes"].append("MOVETO")
        if not curved:
            path["codes"].extend(["LINETO"] * 6)

            # Start
            theta = atan2(*((coordst[1] - coordst[0])[::-1]))
            voff = 0 * coordst[0]
            voff[:] = [cos(theta), sin(theta)]
            voff *= sizes[0] / 2
            start = coordst[0] + voff

            # End with arrow (base-left-top-right-base)
            theta = atan2(*((coordst[1] - coordst[0])[::-1]))
            voff_unity = 0 * coordst[0]
            voff_unity[:] = [cos(theta), sin(theta)]
            voff = voff_unity * sizes[1] / 2
            tip = coordst[1] - voff

            voff_unity_90 = voff_unity @ [[0, 1], [-1, 0]]
            headbase = tip - arrow_size * voff_unity
            headleft = headbase + 0.5 * arrow_width * voff_unity_90
            headright = headbase - 0.5 * arrow_width * voff_unity_90
            # This is a dirty trick to make the facecolor work
            # without making a separate Patch, which would be a little messy
            path["vertices"].extend(
                [
                    headbase,
                    start,
                    headbase,
                    headleft,
                    tip,
                    headright,
                    headbase,
                ]
            )
        else:
            # Bezier
            aux1, aux2 = get_bezier_control_points_for_curved_edge(
                *coordst.ravel(),
                curved,
            )

            # Start
            theta = atan2(*((aux1 - coordst[0])[::-1]))
            voff_unity = 0 * coordst[0]
            voff_unity[:] = [cos(theta), sin(theta)]
            start = coordst[0] + voff_unity * sizes[0] / 2

            # End with arrow (base-left-top-right-base)
            theta = atan2(*((coordst[1] - aux2)[::-1]))
            voff_unity = 0 * coordst[0]
            voff_unity[:] = [cos(theta), sin(theta)]
            voff_unity_90 = voff_unity @ [[0, 1], [-1, 0]]
            tip = coordst[1] - voff_unity * sizes[1] / 2
            headbase = tip - arrow_size * voff_unity
            headleft = headbase + 0.5 * arrow_width * voff_unity_90
            headright = headbase - 0.5 * arrow_width * voff_unity_90

            # This is a dirty trick to make the facecolor work
            # without making a separate Patch, which would be a little messy
            path["codes"].extend(["CURVE4"] * 6 + ["LINETO"] * 4)
            path["vertices"].extend(
                [
                    headbase,
                    aux2,
                    aux1,
                    start,
                    aux1,
                    aux2,
                    headbase,
                    headleft,
                    tip,
                    headright,
                    headbase,
                ]
            )

        path = mpl.path.Path(
            path["vertices"],
            codes=[getattr(mpl.path.Path, x) for x in path["codes"]],
        )
        path.vertices = trans_inv(path.vertices)
        return path

    def draw(self, renderer):
        if self._visual_vertices is not None:
            self._paths = self._compute_paths()
        return super().draw(renderer)

    def get_arrow_sizes(self):
        """Same as get_arrow_size."""
        return self.get_arrow_size()

    def get_arrow_size(self):
        """Get arrow sizes for the edges (directed only).

        @return: An array of arrow sizes.
        """
        import numpy as np

        if self._arrow_sizes is None:
            arrow_sizes = [0 for x in self.get_paths()]
        else:
            arrow_sizes = self._arrow_sizes
        return np.array(arrow_sizes)

    def set_arrow_size(self, sizes):
        """Set arrow sizes.

        @param sizes: A sequence of arrow sizes or a single size.
        """
        try:
            iter(sizes)
        except TypeError:
            sizes = [sizes] * len(self._paths)
        self._arrow_sizes = sizes
        self.stale = True

    def set_arrow_sizes(self, sizes):
        """Same as set_arrow_size"""
        return self.set_arrow_size(sizes)

    def get_arrow_widths(self):
        """Same as get_arrow_width."""
        return self.get_arrow_width()

    def get_arrow_width(self):
        """Get arrow widths for the edges (directed only).

        @return: An array of arrow widths.
        """
        import numpy as np

        if self._arrow_widths is None:
            arrow_widths = [0 for x in self.get_paths()]
        else:
            arrow_widths = self._arrow_widths
        return np.array(arrow_widths)

    def set_arrow_width(self, widths):
        """Set arrow widths.

        @param widths: A sequence of arrow widths or a single width.
        """
        try:
            iter(widths)
        except TypeError:
            widths = [widths] * len(self._paths)
        self._arrow_widths = widths
        self.stale = True

    def set_arrow_widths(self, widths):
        """Same as set_arrow_width"""
        return self.set_arrow_width(widths)

    @property
    def stale(self):
        return super().stale

    @stale.setter
    def stale(self, val):
        PatchCollection.stale.fset(self, val)
        if val and hasattr(self, "stale_callback_post"):
            self.stale_callback_post(self)



"""
.. _tutorials-iplotx:

==============================
Visualising graphs with iplotx
==============================
``iplotx`` (https://iplotx.readthedocs.io) is a library for visualisation of graphs/networks
with direct compatibility with both igraph and NetworkX. It uses ``matplotlib`` behind the
scenes so the results are compatible with the current igraph matplotlib backend and many
additional chart types (e.g. bar charts, annotations).

Compared to the standard visualisations shipped with igraph, ``iplotx`` offers:

- More styling options
- More consistent behaviour across DPI resolutions and backends
- More consistent matplotlib artists for plot editing and animation

"""

import igraph as ig
import iplotx as ipx

# Construct a graph with 5 vertices
n_vertices = 5
edges = [(0, 1), (0, 2), (0, 3), (0, 4), (1, 2), (1, 3), (1, 4), (3, 4)]
g = ig.Graph(n_vertices, edges)

# Set attributes for the graph, nodes, and edges
g["title"] = "Small Social Network"
g.vs["name"] = [
    "Daniel Morillas",
    "Kathy Archer",
    "Kyle Ding",
    "Joshua Walton",
    "Jana Hoyer",
]
g.vs["gender"] = ["M", "F", "F", "M", "F"]
g.es["married"] = [False, False, False, False, False, False, False, True]

# Set individual attributes
g.vs[1]["name"] = "Kathy Morillas"
g.es[0]["married"] = True

# Plot using iplotx
ipx.network(
    g,
    layout="circle",  # print nodes in a circular layout
    vertex_marker="s",
    vertex_size=45,
    vertex_linewidth=2,
    vertex_facecolor=[
        "lightblue" if gender == "M" else "deeppink" for gender in g.vs["gender"]
    ],
    vertex_label_color=[
        "black" if gender == "M" else "white" for gender in g.vs["gender"]
    ],
    vertex_edgecolor="black",
    vertex_labels=[name.replace(" ", "\n") for name in g.vs["name"]],
    edge_linewidth=[2 if married else 1 for married in g.es["married"]],
    edge_color=["#7142cf" if married else "#AAA" for married in g.es["married"]],
    edge_padding=3,
    aspect=1.0,
)



import random
import unittest

from igraph import Graph, InternalError


class RandomWalkTests(unittest.TestCase):
    def setUp(self):
        random.seed(42)

    def validate_walk(self, g, walk, start, length, mode="out"):
        self.assertEqual(len(walk), length + 1)

        prev = None
        for vertex in walk:
            if prev is not None:
                self.assertTrue(vertex in g.neighbors(prev, mode=mode))
            else:
                self.assertEqual(start, vertex)
            prev = vertex

    def validate_edge_walk(self, g, walk, start, length, mode="out"):
        self.assertEqual(len(walk), length)

        prev_vertices = None
        for edgeid in walk:
            vertices = g.es[edgeid].tuple
            if prev_vertices is not None:
                self.assertTrue(
                    vertices[0] in prev_vertices or vertices[1] in prev_vertices
                )
            else:
                self.assertTrue(start in vertices)
            prev_vertices = vertices

    def testRandomWalkUndirected(self):
        g = Graph.GRG(100, 0.2)
        for _i in range(100):
            start = random.randint(0, g.vcount() - 1)
            length = random.randint(0, 10)
            walk = g.random_walk(start, length)
            self.validate_walk(g, walk, start, length)

    def testRandomWalkDirectedOut(self):
        g = Graph.Tree(121, 3, mode="out")
        mode = "out"
        for _i in range(100):
            start = 0
            length = random.randint(0, 4)
            walk = g.random_walk(start, length, mode)
            self.validate_walk(g, walk, start, length, mode)

    def testRandomWalkDirectedIn(self):
        g = Graph.Tree(121, 3, mode="out")
        mode = "in"
        for _i in range(100):
            start = random.randint(40, g.vcount() - 1)
            length = random.randint(0, 4)
            walk = g.random_walk(start, length, mode)
            self.validate_walk(g, walk, start, length, mode)

    def testRandomWalkDirectedAll(self):
        g = Graph.Tree(121, 3, mode="out")
        mode = "all"
        for _i in range(100):
            start = random.randint(0, g.vcount() - 1)
            length = random.randint(0, 10)
            walk = g.random_walk(start, length, mode)
            self.validate_walk(g, walk, start, length, mode)

    def testRandomWalkStuck(self):
        g = Graph.Ring(10, circular=False, directed=True)
        walk = g.random_walk(5, 20)
        self.assertEqual([5, 6, 7, 8, 9], walk)
        self.assertRaises(InternalError, g.random_walk, 5, 20, stuck="error")

    def testRandomWalkUndirectedVertices(self):
        g = Graph.GRG(100, 0.2)
        for _i in range(10):
            start = random.randint(0, g.vcount() - 1)
            length = random.randint(0, 10)
            walk = g.random_walk(start, length, return_type="vertices")
            self.validate_walk(g, walk, start, length)

    def testRandomWalkUndirectedEdges(self):
        g = Graph.GRG(100, 0.2)
        for _i in range(10):
            start = random.randint(0, g.vcount() - 1)
            length = random.randint(0, 10)
            walk = g.random_walk(start, length, return_type="edges")
            self.validate_edge_walk(g, walk, start, length)

    def testRandomWalkUndirectedBoth(self):
        g = Graph.GRG(100, 0.2)
        for _i in range(10):
            start = random.randint(0, g.vcount() - 1)
            length = random.randint(0, 10)
            walk_dic = g.random_walk(start, length, return_type="both")
            self.assertTrue("vertices" in walk_dic)
            self.assertTrue("edges" in walk_dic)
            self.validate_edge_walk(g, walk_dic["edges"], start, length)
            self.validate_walk(g, walk_dic["vertices"], start, length)

    def testRandomWalkUndirectedWeighted(self):
        g = Graph.GRG(100, 0.2)
        g.es["weight"] = [1.0 for i in range(g.ecount())]
        for _i in range(100):
            start = random.randint(0, g.vcount() - 1)
            length = random.randint(0, 10)
            walk = g.random_walk(start, length, weights="weight")
            self.validate_walk(g, walk, start, length)


def suite():
    random_walk_suite = unittest.defaultTestLoader.loadTestsFromTestCase(
        RandomWalkTests
    )
    return unittest.TestSuite([random_walk_suite])


def test():
    runner = unittest.TextTestRunner()
    runner.run(suite())


if __name__ == "__main__":
    test()



"""
Utility classes for drawing routines.
"""

from collections import defaultdict
from math import atan2, cos, hypot, sin
from typing import NamedTuple

from igraph.utils import consecutive_pairs

__all__ = (
    "BoundingBox",
    "Point",
    "Rectangle",
    "calculate_corner_radii",
    "euclidean_distance",
    "evaluate_cubic_bezier",
    "get_bezier_control_points_for_curved_edge",
    "intersect_bezier_curve_and_circle",
    "str_to_orientation",
    "autocurve",
)

#####################################################################


class Rectangle:
    """Class representing a rectangle."""

    __slots__ = ("_left", "_top", "_right", "_bottom")

    def __init__(self, *args):
        """Creates a rectangle.

        The corners of the rectangle can be specified by either a tuple
        (four items, two for each corner, respectively), four separate numbers
        (X and Y coordinates for each corner) or two separate numbers (width
        and height, the upper left corner is assumed to be at (0,0))"""
        coords = None
        if len(args) == 1:
            if isinstance(args[0], Rectangle):
                coords = args[0].coords
            elif len(args[0]) >= 4:
                coords = tuple(args[0])[0:4]
            elif len(args[0]) == 2:
                coords = (0, 0, args[0][0], args[0][1])
        elif len(args) == 4:
            coords = tuple(args)
        elif len(args) == 2:
            coords = (0, 0, args[0], args[1])
        if coords is None:
            raise ValueError("invalid coordinate format")

        try:
            coords = tuple(float(coord) for coord in coords)
        except ValueError:
            raise ValueError("invalid coordinate format, numbers expected") from None

        self.coords = coords

    @property
    def coords(self):
        """The coordinates of the corners.

        The coordinates are returned as a 4-tuple in the following order:
        left edge, top edge, right edge, bottom edge.
        """
        return self._left, self._top, self._right, self._bottom

    @coords.setter
    def coords(self, coords):
        """Sets the coordinates of the corners.

        @param coords: a 4-tuple with the coordinates of the corners
        """
        self._left, self._top, self._right, self._bottom = coords
        if self._left > self._right:
            self._left, self._right = self._right, self._left
        if self._top > self._bottom:
            self._bottom, self._top = self._top, self._bottom

    @property
    def width(self):
        """The width of the rectangle"""
        return self._right - self._left

    @width.setter
    def width(self, value):
        """Sets the width of the rectangle by adjusting the right edge."""
        self._right = self._left + value

    @property
    def height(self):
        """The height of the rectangle"""
        return self._bottom - self._top

    @height.setter
    def height(self, value):
        """Sets the height of the rectangle by adjusting the bottom edge."""
        self._bottom = self._top + value

    @property
    def left(self):
        """The X coordinate of the left side of the box"""
        return self._left

    @left.setter
    def left(self, value):
        """Sets the X coordinate of the left side of the box"""
        self._left = float(value)
        self._right = max(self._left, self._right)

    @property
    def right(self):
        """The X coordinate of the right side of the box"""
        return self._right

    @right.setter
    def right(self, value):
        """Sets the X coordinate of the right side of the box"""
        self._right = float(value)
        self._left = min(self._left, self._right)

    @property
    def top(self):
        """The Y coordinate of the top edge of the box"""
        return self._top

    @top.setter
    def top(self, value):
        """Sets the Y coordinate of the top edge of the box"""
        self._top = value
        self._bottom = max(self._bottom, self._top)

    @property
    def bottom(self):
        """The Y coordinate of the bottom edge of the box"""
        return self._bottom

    @bottom.setter
    def bottom(self, value):
        """Sets the Y coordinate of the bottom edge of the box"""
        self._bottom = value
        self._top = min(self._bottom, self._top)

    @property
    def midx(self):
        """The X coordinate of the center of the box"""
        return (self._left + self._right) / 2.0

    @midx.setter
    def midx(self, value):
        """Moves the center of the box to the given X coordinate"""
        dx = value - (self._left + self._right) / 2.0
        self._left += dx
        self._right += dx

    @property
    def midy(self):
        """The Y coordinate of the center of the box"""
        return (self._top + self._bottom) / 2.0

    @midy.setter
    def midy(self, value):
        """Moves the center of the box to the given Y coordinate"""
        dy = value - (self._top + self._bottom) / 2.0
        self._top += dy
        self._bottom += dy

    @property
    def shape(self):
        """The shape of the rectangle (width, height)"""
        return self._right - self._left, self._bottom - self._top

    @shape.setter
    def shape(self, shape):
        """Sets the shape of the rectangle (width, height)."""
        self.width, self.height = shape

    def contract(self, margins):
        """Contracts the rectangle by the given margins.

        @return: a new L{Rectangle} object.
        """
        if isinstance(margins, int) or isinstance(margins, float):
            margins = [float(margins)] * 4
        if len(margins) != 4:
            raise ValueError("margins must be a 4-tuple or a single number")
        nx1, ny1 = self._left + margins[0], self._top + margins[1]
        nx2, ny2 = self._right - margins[2], self._bottom - margins[3]
        if nx1 > nx2:
            nx1 = (nx1 + nx2) / 2.0
            nx2 = nx1
        if ny1 > ny2:
            ny1 = (ny1 + ny2) / 2.0
            ny2 = ny1
        return self.__class__(nx1, ny1, nx2, ny2)

    def expand(self, margins):
        """Expands the rectangle by the given margins.

        @return: a new L{Rectangle} object.
        """
        if isinstance(margins, int) or isinstance(margins, float):
            return self.contract(-float(margins))
        return self.contract([-float(margin) for margin in margins])

    def isdisjoint(self, other):
        """Returns C{True} if the two rectangles have no intersection.

        Example::

            >>> r1 = Rectangle(10, 10, 30, 30)
            >>> r2 = Rectangle(20, 20, 50, 50)
            >>> r3 = Rectangle(70, 70, 90, 90)
            >>> r1.isdisjoint(r2)
            False
            >>> r2.isdisjoint(r1)
            False
            >>> r1.isdisjoint(r3)
            True
            >>> r3.isdisjoint(r1)
            True
        """
        return (
            self._left > other._right
            or self._right < other._left
            or self._top > other._bottom
            or self._bottom < other._top
        )

    def isempty(self):
        """Returns C{True} if the rectangle is empty (i.e. it has zero
        width and height).

        Example::

            >>> r1 = Rectangle(10, 10, 30, 30)
            >>> r2 = Rectangle(70, 70, 90, 90)
            >>> r1.isempty()
            False
            >>> r2.isempty()
            False
            >>> r1.intersection(r2).isempty()
            True
        """
        return self._left == self._right and self._top == self._bottom

    def intersection(self, other):
        """Returns the intersection of this rectangle with another.

        Example::

            >>> r1 = Rectangle(10, 10, 30, 30)
            >>> r2 = Rectangle(20, 20, 50, 50)
            >>> r3 = Rectangle(70, 70, 90, 90)
            >>> r1.intersection(r2)
            Rectangle(20.0, 20.0, 30.0, 30.0)
            >>> r2 & r1
            Rectangle(20.0, 20.0, 30.0, 30.0)
            >>> r2.intersection(r1) == r1.intersection(r2)
            True
            >>> r1.intersection(r3)
            Rectangle(0.0, 0.0, 0.0, 0.0)
        """
        if self.isdisjoint(other):
            return Rectangle(0, 0, 0, 0)
        return Rectangle(
            max(self._left, other._left),
            max(self._top, other._top),
            min(self._right, other._right),
            min(self._bottom, other._bottom),
        )

    __and__ = intersection

    def translate(self, dx, dy):
        """Translates the rectangle in-place.

        Example:

            >>> r = Rectangle(10, 20, 50, 70)
            >>> r.translate(30, -10)
            >>> r
            Rectangle(40.0, 10.0, 80.0, 60.0)

        @param dx: the X coordinate of the translation vector
        @param dy: the Y coordinate of the translation vector
        """
        self._left += dx
        self._right += dx
        self._top += dy
        self._bottom += dy

    def union(self, other):
        """Returns the union of this rectangle with another.

        The resulting rectangle is the smallest rectangle that contains both
        rectangles.

        Example::

            >>> r1 = Rectangle(10, 10, 30, 30)
            >>> r2 = Rectangle(20, 20, 50, 50)
            >>> r3 = Rectangle(70, 70, 90, 90)
            >>> r1.union(r2)
            Rectangle(10.0, 10.0, 50.0, 50.0)
            >>> r2 | r1
            Rectangle(10.0, 10.0, 50.0, 50.0)
            >>> r2.union(r1) == r1.union(r2)
            True
            >>> r1.union(r3)
            Rectangle(10.0, 10.0, 90.0, 90.0)
        """
        return Rectangle(
            min(self._left, other._left),
            min(self._top, other._top),
            max(self._right, other._right),
            max(self._bottom, other._bottom),
        )

    __or__ = union

    def __ior__(self, other):
        """Expands this rectangle to include itself and another completely while
        still being as small as possible.

        Example::

            >>> r1 = Rectangle(10, 10, 30, 30)
            >>> r2 = Rectangle(20, 20, 50, 50)
            >>> r3 = Rectangle(70, 70, 90, 90)
            >>> r1 |= r2
            >>> r1
            Rectangle(10.0, 10.0, 50.0, 50.0)
            >>> r1 |= r3
            >>> r1
            Rectangle(10.0, 10.0, 90.0, 90.0)
        """
        self._left = min(self._left, other._left)
        self._top = min(self._top, other._top)
        self._right = max(self._right, other._right)
        self._bottom = max(self._bottom, other._bottom)
        return self

    def __repr__(self):
        return "%s(%s, %s, %s, %s)" % (
            self.__class__.__name__,
            self._left,
            self._top,
            self._right,
            self._bottom,
        )

    def __eq__(self, other):
        return self.coords == other.coords

    def __ne__(self, other):
        return self.coords != other.coords

    def __bool__(self):
        return self._left != self._right or self._top != self._bottom

    def __hash__(self):
        return hash(self.coords)


#####################################################################


class BoundingBox(Rectangle):
    """Class representing a bounding box (a rectangular area) that
    encloses some objects."""

    def __ior__(self, other):
        """Replaces this bounding box with the union of itself and
        another.

        Example::

            >>> box1 = BoundingBox(10, 20, 50, 60)
            >>> box2 = BoundingBox(70, 40, 100, 90)
            >>> box1 |= box2
            >>> print(box1)
            BoundingBox(10.0, 20.0, 100.0, 90.0)
        """
        self._left = min(self._left, other._left)
        self._top = min(self._top, other._top)
        self._right = max(self._right, other._right)
        self._bottom = max(self._bottom, other._bottom)
        return self

    def __or__(self, other):
        """Takes the union of this bounding box with another.

        The result is a bounding box which encloses both bounding
        boxes.

        Example::

            >>> box1 = BoundingBox(10, 20, 50, 60)
            >>> box2 = BoundingBox(70, 40, 100, 90)
            >>> box1 | box2
            BoundingBox(10.0, 20.0, 100.0, 90.0)
        """
        return self.__class__(
            min(self._left, other._left),
            min(self._top, other._top),
            max(self._right, other._right),
            max(self._bottom, other._bottom),
        )


#####################################################################


class FakeModule:
    """Fake module that raises an exception for everything"""

    def __init__(self, message):
        """Constructor.

        @param message: message to print in exceptions raised from this module
        """
        self._message = message

    def __getattr__(self, _):
        raise AttributeError(self._message)

    def __call__(self, _):
        raise TypeError(self._message)

    def __setattr__(self, key, value):
        if key == "_message":
            super().__setattr__(key, value)
        else:
            raise AttributeError(self._message)


#####################################################################


class Point(NamedTuple("_Point", [("x", float), ("y", float)])):
    """Class representing a point on the 2D plane."""

    def __add__(self, other):
        """Adds the coordinates of a point to another one"""
        return self.__class__(x=self.x + other.x, y=self.y + other.y)

    def __sub__(self, other):
        """Subtracts the coordinates of a point to another one"""
        return self.__class__(x=self.x - other.x, y=self.y - other.y)

    def __mul__(self, scalar):
        """Multiplies the coordinates by a scalar"""
        return self.__class__(x=self.x * scalar, y=self.y * scalar)

    __rmul__ = __mul__

    def __div__(self, scalar):
        """Divides the coordinates by a scalar"""
        return self.__class__(x=self.x / scalar, y=self.y / scalar)

    def as_polar(self):
        """Returns the polar coordinate representation of the point.

        @return: the radius and the angle in a tuple.
        """
        return len(self), atan2(self.y, self.x)

    def distance(self, other):
        """Returns the distance of the point from another one.

        Example:

            >>> p1 = Point(5, 7)
            >>> p2 = Point(8, 3)
            >>> p1.distance(p2)
            5.0
        """
        dx, dy = self.x - other.x, self.y - other.y
        return (dx * dx + dy * dy) ** 0.5

    def interpolate(self, other, ratio=0.5):
        """Linearly interpolates between the coordinates of this point and
        another one.

        @param  other:  the other point
        @param  ratio:  the interpolation ratio between 0 and 1. Zero will
          return this point, 1 will return the other point.
        """
        ratio = float(ratio)
        return self.__class__(
            x=self.x * (1.0 - ratio) + other.x * ratio,
            y=self.y * (1.0 - ratio) + other.y * ratio,
        )

    def length(self):
        """Returns the length of the vector pointing from the origin to this
        point."""
        return (self.x**2 + self.y**2) ** 0.5

    def normalized(self):
        """Normalizes the coordinates of the point s.t. its length will be 1
        after normalization. Returns the normalized point."""
        len = self.length()
        if len == 0:
            return self.__class__(x=self.x, y=self.y)
        return self.__class__(x=self.x / len, y=self.y / len)

    def sq_length(self):
        """Returns the squared length of the vector pointing from the origin
        to this point."""
        return self.x**2 + self.y**2

    def towards(self, other, distance=0):
        """Returns the point that is at a given distance from this point
        towards another one."""
        if not distance:
            return self

        angle = atan2(other.y - self.y, other.x - self.x)
        return self.__class__(
            self.x + distance * cos(angle), self.y + distance * sin(angle)
        )

    @classmethod
    def FromPolar(cls, radius, angle):
        """Constructs a point from polar coordinates.

        C{radius} is the distance of the point from the origin; C{angle} is the
        angle between the X axis and the vector pointing to the point from
        the origin.
        """
        return cls(radius * cos(angle), radius * sin(angle))


def calculate_corner_radii(points, corner_radius):
    """Given a list of points and a desired corner radius, returns a list
    containing proposed corner radii for each of the points such that it is
    ensured that the corner radius at a point is never larger than half of
    the minimum distance between the point and its neighbors.
    """
    points = [Point(*point) for point in points]
    side_vecs = [v - u for u, v in consecutive_pairs(points, circular=True)]
    half_side_lengths = [side.length() / 2 for side in side_vecs]
    corner_radii = [corner_radius] * len(points)
    for idx in range(len(corner_radii)):
        prev_idx = -1 if idx == 0 else idx - 1
        radii = [corner_radius, half_side_lengths[prev_idx], half_side_lengths[idx]]
        corner_radii[idx] = min(radii)
    return corner_radii


def euclidean_distance(x1, y1, x2, y2):
    """Computes the Euclidean distance between points (x1,y1) and (x2,y2)."""
    return hypot(x2 - x1, y2 - y1)


def evaluate_cubic_bezier(x0, y0, x1, y1, x2, y2, x3, y3, t):
    """Evaluates the Bezier curve from point (x0,y0) to (x3,y3)
    via control points (x1,y1) and (x2,y2) at t. t is typically in the range
    [0; 1] such that 0 returns (x0, y0) and 1 returns (x3, y3).
    """
    xt = (
        (1.0 - t) ** 3 * x0
        + 3.0 * t * (1.0 - t) ** 2 * x1
        + 3.0 * t**2 * (1.0 - t) * x2
        + t**3 * x3
    )
    yt = (
        (1.0 - t) ** 3 * y0
        + 3.0 * t * (1.0 - t) ** 2 * y1
        + 3.0 * t**2 * (1.0 - t) * y2
        + t**3 * y3
    )
    return xt, yt


def get_bezier_control_points_for_curved_edge(x1, y1, x2, y2, curvature):
    """Helper function that calculates the Bezier control points for a
    curved edge that goes from (x1, y1) to (x2, y2).
    """
    aux1 = (
        (2 * x1 + x2) / 3.0 - curvature * 0.5 * (y2 - y1),
        (2 * y1 + y2) / 3.0 + curvature * 0.5 * (x2 - x1),
    )
    aux2 = (
        (x1 + 2 * x2) / 3.0 - curvature * 0.5 * (y2 - y1),
        (y1 + 2 * y2) / 3.0 + curvature * 0.5 * (x2 - x1),
    )
    return aux1, aux2


def intersect_bezier_curve_and_circle(
    x0, y0, x1, y1, x2, y2, x3, y3, radius, max_iter=10
):
    """Binary search solver for finding the intersection of a Bezier curve
    and a circle centered at the curve's end point.

    Returns the x, y coordinates of the intersection point.
    """
    # The exact formulation of the problem is a quartic equation and it is
    # probably not worth coding up an exact quartic solver. The solution below
    # uses binary search. Another solution would be simply to intersect the
    # circle with the line pointing from (x2, y2) to (x3, y3) as the difference
    # is likely to be negligible.

    precision = radius / 20.0
    source_target_distance = euclidean_distance(x0, y0, x3, y3)
    radius = float(radius)
    t0 = 1.0
    t1 = 1.0 - radius / source_target_distance

    xt1, yt1 = evaluate_cubic_bezier(x0, y0, x1, y1, x2, y2, x3, y3, t1)

    distance_t0 = 0
    distance_t1 = euclidean_distance(x3, y3, xt1, yt1)
    counter = 0
    while abs(distance_t1 - radius) > precision and counter < max_iter:
        if ((distance_t1 - radius) > 0) != ((distance_t0 - radius) > 0):
            t_new = (t0 + t1) / 2.0
        else:
            if abs(distance_t1 - radius) < abs(distance_t0 - radius):
                # If t1 gets us closer to the circumference step in the
                # same direction
                t_new = t1 + (t1 - t0) / 2.0
            else:
                t_new = t1 - (t1 - t0)
        t_new = 1 if t_new > 1 else (0 if t_new < 0 else t_new)
        t0, t1 = t1, t_new
        distance_t0 = distance_t1
        xt1, yt1 = evaluate_cubic_bezier(x0, y0, x1, y1, x2, y2, x3, y3, t1)
        distance_t1 = euclidean_distance(x3, y3, xt1, yt1)
        counter += 1

    return evaluate_cubic_bezier(x0, y0, x1, y1, x2, y2, x3, y3, t1)


def str_to_orientation(value, reversed_horizontal=False, reversed_vertical=False):
    """Tries to interpret a string as an orientation value.

    The following basic values are understood: ``left-right``, ``bottom-top``,
    ``right-left``, ``top-bottom``. Possible aliases are:

      - ``horizontal``, ``horiz``, ``h`` and ``lr`` for ``left-right``

      - ``vertical``, ``vert``, ``v`` and ``tb`` for top-bottom.

      - ``lr`` for ``left-right``.

      - ``rl`` for ``right-left``.

    ``reversed_horizontal`` reverses the meaning of ``horizontal``, ``horiz``
    and ``h`` to ``rl`` (instead of ``lr``); similarly, ``reversed_vertical``
    reverses the meaning of ``vertical``, ``vert`` and ``v`` to ``bt``
    (instead of ``tb``).

    Returns one of ``lr``, ``rl``, ``tb`` or ``bt``, or throws ``ValueError``
    if the string cannot be interpreted as an orientation.
    """

    aliases = {
        "left-right": "lr",
        "right-left": "rl",
        "top-bottom": "tb",
        "bottom-top": "bt",
        "top-down": "tb",
        "bottom-up": "bt",
        "td": "tb",
        "bu": "bt",
    }

    dir = ["lr", "rl"][reversed_horizontal]
    aliases.update(horizontal=dir, horiz=dir, h=dir)

    dir = ["tb", "bt"][reversed_vertical]
    aliases.update(vertical=dir, vert=dir, v=dir)

    result = aliases.get(value, value)
    if result not in ("lr", "rl", "tb", "bt"):
        raise ValueError("unknown orientation: %s" % result)
    return result


def autocurve(graph, attribute="curved", default=0):
    """Calculates curvature values for each of the edges in the graph to make
    sure that multiple edges are shown properly on a graph plot.

    This function checks the multiplicity of each edge in the graph and
    assigns curvature values (numbers between -1 and 1, corresponding to
    CCW (-1), straight (0) and CW (1) curved edges) to them. The assigned
    values are either stored in an edge attribute or returned as a list,
    depending on the value of the I{attribute} argument.

    @param graph: the graph on which the calculation will be run
    @param attribute: the name of the edge attribute to save the curvature
      values to. The default value is C{curved}, which is the name of the
      edge attribute the default graph plotter checks to decide whether an
      edge should be curved on the plot or not. If I{attribute} is C{None},
      the result will not be stored.
    @param default: the default curvature for single edges. Zero means that
      single edges will be straight. If you want single edges to be curved
      as well, try passing 0.5 or -0.5 here.
    @return: the list of curvature values if I{attribute} is C{None},
      otherwise C{None}.
    """

    # The following loop could be re-written in C if it turns out to be a
    # bottleneck. Unfortunately we cannot use Graph.count_multiple() here
    # because we have to ignore edge directions.
    multiplicities = defaultdict(list)
    for edge in graph.es:
        u, v = edge.tuple
        if u > v:
            multiplicities[v, u].append(edge.index)
        else:
            multiplicities[u, v].append(edge.index)

    result = [default] * graph.ecount()
    for eids in multiplicities.values():
        # Is it a single edge?
        if len(eids) < 2:
            continue

        if len(eids) % 2 == 1:
            # Odd number of edges; the last will be straight
            result[eids.pop()] = 0

        # Arrange the remaining edges
        curve = 2.0 / (len(eids) + 2)
        dcurve, sign = curve, 1
        for idx, eid in enumerate(eids):
            edge = graph.es[eid]
            if edge.source > edge.target:
                result[eid] = -sign * curve
            else:
                result[eid] = sign * curve
            if idx % 2 == 1:
                curve += dcurve
            sign *= -1

    if attribute is None:
        return result

    graph.es[attribute] = result



# vim:ts=4:sw=4:sts=4:et
# -*- coding: utf-8 -*-
"""Classes representing matchings on graphs."""

from igraph._igraph import Vertex


class Matching:
    """A matching of vertices in a graph.

    A matching of an undirected graph is a set of edges such that each
    vertex is incident on at most one matched edge. When each vertex is
    incident on I{exactly} one matched edge, the matching called
    I{perfect}. This class is used in C{igraph} to represent non-perfect
    and perfect matchings in undirected graphs.

    This class is usually not instantiated directly, everything
    is taken care of by the functions that return matchings.

    Examples:

      >>> from igraph import Graph
      >>> g = Graph.Famous("noperfectmatching")
      >>> matching = g.maximum_matching()
    """

    def __init__(self, graph, matching, types=None):
        """Initializes the matching.

        @param graph: the graph the matching belongs to
        @param matching: a numeric vector where element I{i} corresponds to
          vertex I{i} of the graph. Element I{i} is -1 or if the corresponding
          vertex is unmatched, otherwise it contains the index of the vertex to
          which vertex I{i} is matched.
        @param types: the types of the vertices if the graph is bipartite.
          It must either be the name of a vertex attribute (which will be
          retrieved for all vertices) or a list. Elements in the list will be
          converted to boolean values C{True} or C{False}, and this will
          determine which part of the bipartite graph a given vertex belongs to.
        @raise ValueError: if the matching vector supplied does not describe
          a valid matching of the graph.
        """
        self._graph = graph
        self._matching = None
        self._num_matched = 0
        self._types = None

        if isinstance(types, str):
            types = graph.vs[types]

        self.types = types
        self.matching = matching

    def __len__(self):
        return self._num_matched

    def __repr__(self):
        if self._types is not None:
            return "%s(%r,%r,types=%r)" % (
                self.__class__.__name__,
                self._graph,
                self._matching,
                self._types,
            )
        else:
            return "%s(%r,%r)" % (self.__class__.__name__, self._graph, self._matching)

    def __str__(self):
        if self._types is not None:
            return "Bipartite graph matching (%d matched vertex pairs)" % len(self)
        else:
            return "Graph matching (%d matched vertex pairs)" % len(self)

    def edges(self):
        """Returns an edge sequence that contains the edges in the matching.

        If there are multiple edges between a pair of matched vertices, only one
        of them will be returned.
        """
        get_eid = self._graph.get_eid
        eidxs = [
            get_eid(u, v, directed=False)
            for u, v in enumerate(self._matching)
            if v != -1 and u <= v
        ]
        return self._graph.es[eidxs]

    @property
    def graph(self):
        """Returns the graph corresponding to the matching."""
        return self._graph

    def is_maximal(self):
        """Returns whether the matching is maximal.

        A matching is maximal when it is not possible to extend it any more
        with extra edges; in other words, unmatched vertices in the graph
        must be adjacent to matched vertices only.
        """
        return self._graph._is_maximal_matching(self._matching, types=self._types)

    def is_matched(self, vertex):
        """Returns whether the given vertex is matched to another one."""
        if isinstance(vertex, Vertex):
            vertex = vertex.index
        return self._matching[vertex] >= 0

    def match_of(self, vertex):
        """Returns the vertex a given vertex is matched to.

        @param vertex: the vertex we are interested in; either an integer index
          or an instance of L{Vertex}.
        @return: the index of the vertex matched to the given vertex, either as
          an integer index (if I{vertex} was integer) or as an instance of
          L{Vertex}. When the vertex is unmatched, returns C{None}.
        """
        if isinstance(vertex, Vertex):
            matched = self._matching[vertex.index]
            if matched < 0:
                return None
            return self._graph.vs[matched]

        matched = self._matching[vertex]
        if matched < 0:
            return None
        return matched

    @property
    def matching(self):
        """Returns the matching vector where element I{i} contains the ID of
        the vertex that vertex I{i} is matched to.

        The matching vector will contain C{-1} for unmatched vertices.
        """
        return self._matching

    @matching.setter
    def matching(self, value):
        """Sets the matching vector.

        @param value: the matching vector which must contain the ID of the
          vertex matching vertex I{i} at the I{i}th position, or C{-1} if
          the vertex is unmatched.
        @raise ValueError: if the matching vector supplied does not describe
          a valid matching of the graph.
        """
        if not self.graph._is_matching(value, types=self._types):
            raise ValueError("not a valid matching")
        self._matching = list(value)
        self._num_matched = sum(1 for i in self._matching if i >= 0) // 2

    @property
    def types(self):
        """Returns the type vector if the graph is bipartite.

        Element I{i} of the type vector will be C{False} or C{True} depending
        on which side of the bipartite graph vertex I{i} belongs to.

        For non-bipartite graphs, this property returns C{None}.
        """
        return self._types

    @types.setter
    def types(self, value):
        types = [bool(x) for x in value]
        if len(types) < self._graph.vcount():
            raise ValueError("type vector too short")
        self._types = types



"""
.. _tutorials-bipartite-matching:

==========================
Maximum Bipartite Matching
==========================

This example demonstrates an efficient way to find and visualise a maximum biparite matching using :meth:`igraph.Graph.maximum_bipartite_matching`.
"""

import igraph as ig
import matplotlib.pyplot as plt

# %%
# First, we construct a bipartite graph, assigning:
#  - nodes 0-4 to one side
#  - nodes 5-8 to the other side
g = ig.Graph.Bipartite(
    [0, 0, 0, 0, 0, 1, 1, 1, 1],
    [(0, 5), (1, 6), (1, 7), (2, 5), (2, 8), (3, 6), (4, 5), (4, 6)],
)

# %%
# We can easily check that the graph is indeed bipartite:
assert g.is_bipartite()

# %%
# Now can can compute the maximum bipartite matching:
matching = g.maximum_bipartite_matching()

# %%
# It's easy to print matching pairs of vertices
matching_size = 0
print("Matching is:")
for i in range(5):
    print(f"{i} - {matching.match_of(i)}")
    if matching.is_matched(i):
        matching_size += 1
print("Size of maximum matching is:", matching_size)

# %%
# Finally, we can plot the bipartite graph, highlighting the edges connecting
# maximal matches by a red color:
fig, ax = plt.subplots(figsize=(7, 3))
ig.plot(
    g,
    target=ax,
    layout=g.layout_bipartite(),
    vertex_size=30,
    vertex_label=range(g.vcount()),
    vertex_color="lightblue",
    edge_width=[3 if e.target == matching.match_of(e.source) else 1.0 for e in g.es],
    edge_color=["red" if e.target == matching.match_of(e.source) else "black" for e in g.es]
)
